import numpy as np

from tigramite.toymodels import structural_causal_processes as toys


def _nb_latent_before(node, observed_context_indices, node_classification):
    return len(
        [el for el in range(node) if not (el in observed_context_indices or node_classification[el] == "system")])

def _do_dummy_projection(links, node_classification, observed_context_indices, time_dummy_index, space_dummy_index):
    """
    Helper function to augment the true_parents by remove context-context links and adding dummy
    (i.e. perform dummy projection)

    links : dictionary
        Ground truth links
    node_classification : dictionary
        Corresponds to ground truth links
    """

    # remove latent links, shift remaining, add dummy
    augmented_links = {}
    for node in node_classification.keys():
        if node_classification[node] == "system":
            keep_parents = []
            for parent in links[node]:
                if node_classification[parent[0][0]] == "system":
                    keep_parents.append(parent)
                elif node_classification[parent[0][0]] == "time_context":
                    if parent[0][0] in observed_context_indices:
                        keep_parents.append(((parent[0][0] - _nb_latent_before(parent[0][0], observed_context_indices,
                                                                               node_classification),
                                              parent[0][1]), parent[1], parent[2]))
                    else:
                        keep_parents.append(((time_dummy_index, 0), 1., "dummy"))
                elif node_classification[parent[0][0]] == "space_context":
                    if parent[0][0] in observed_context_indices:
                        keep_parents.append(((parent[0][0] - _nb_latent_before(parent[0][0], observed_context_indices,
                                                                               node_classification), parent[0][1]),
                                             parent[1], parent[2]))
                    else:
                        keep_parents.append(((space_dummy_index, 0), 1., "dummy"))
                augmented_links[node] = list(dict.fromkeys(keep_parents))

        # remove all parents of context nodes
        elif node_classification[node] == "time_context":
            if node in observed_context_indices:
                augmented_links[node - _nb_latent_before(node, observed_context_indices, node_classification)] = []
        elif node_classification[node] == "space_context":
            if node in observed_context_indices:
                augmented_links[node - _nb_latent_before(node, observed_context_indices, node_classification)] = []

    augmented_links[time_dummy_index] = []
    augmented_links[space_dummy_index] = []
    return augmented_links

def _shift_link_entries(links, const):
    """
    Helper Function to shift keys and values of a link dictionary by an integer constant.
    """
    shifted_links = {}
    for key in links.keys():
        shifted_key = key + const
        values = links[key]
        shifted_values = [((item + const, lag), c, f) for ((item, lag), c, f) in values]
        shifted_links[shifted_key] = shifted_values
    return shifted_links

def _group_links(links, node_types, node_type):
    return {i: links[i] for i in links.keys() if node_types[i] == node_type}


class ContextModel:
    """Allows to sample from a joint structural causal model over different spatial
        and temporal contexts. Restricts temporal and spatial context nodes to be constant over datasets or
        time, respectively.

        Parameters
        ----------
        links : dict
            Dictionary of format: {0:[((i, -tau), coeff, func),...], 1:[...],
            ...} for all variables where i must be in [0..N-1] and tau >= 0 with
            number of variables N. coeff must be a float and func a python
            callable of one argument.
        node_classification : dictionary
            Classification of nodes into system, or context nodes.
            Keys of the dictionary are from {0, ..., N-1} where N is the number of nodes.
            Options for the values are "system", "time_context", "space_context". The temporal context variables are
            assumed exogenous to the system variables. They cannot interact with the spatial context variables due
            to the assumption that they are constant across datasets.
        transient_fraction : float
            Added percentage of T used as a transient. In total a realization of length
            (transient_fraction + 1)*T will be generated, but then transient_fraction*T will be
            cut off.
        noises : list of callables or list of arrays, optional (default: None)
            Random distribution function that is called with noises[j](T) for system and time-context nodes, or
            noises[j](M) for space-context nodes where M is the number of datasets. If list of arrays,
            for noises corresponding to time-context and system variables the array needs to be of
            shape ((transient_fraction + 1)*T, ), for space-context variables it needs to be of shape (M, )
        seed : int, optional (default: None)
            Random seed.

        Attributes
        -------
        links_tc : dict
            Dictionary of format: {0:[((i, -tau), coeff, func),...], 1:[...],
            ...} for all temporal context variables where i must be in [0..N-1] and tau >= 0 with
            number of variables N. coeff must be a float and func a python
            callable of one argument. The temporal context variables are assumed exogenous
            to the system variables. They cannot interact with the spatial context variables due to the assumption
            that they are constant across datasets.
        links_sc : dict
            Dictionary of format: {0:[((i, -tau), coeff, func),...], 1:[...],
            ...} for all spatial context variables where i must be in [0..N-1] and tau >= 0 with
            number of variables N. coeff must be a float and func a python
            callable of one argument. The spatial context variables are assumed exogenous
            to the system variables. They cannot interact with the temporal context variables due to the assumption
            that they are time-independent, i.e. constant across time.
        links_sys : dict
            Dictionary of format: {0:[((i, -tau), coeff, func),...], 1:[...],
            ...} for all system variables where i must be in [0..N-1] and tau >= 0 with
            number of variables N. coeff must be a float and func a python
            callable of one argument.
        transient_fraction : float
            Added percentage of T used as a transient. In total a realization of length
            (transient_fraction + 1)*T will be generated, but then transient_fraction*T will be
            cut off.
        noises : list of callables or list of arrays, optional (default: None)
            Random distribution function that is called with noises[j](T) for system and time-context nodes, or
            noises[j](M) for space-context nodes where M is the number of datasets. If list of arrays,
            for noises corresponding to time-context and system variables the array needs to be of
            shape ((transient_fraction + 1)*T, ), for space-context variables it needs to be of shape (M, )
        seed : int, optional (default: None)
            Random seed.

        """

    def __init__(self, links={}, node_classification={}, transient_fraction=0.2, noises=None, seed=None):
        self.links_tc = _group_links(links, node_classification, "time_context")
        self.links_sc = _group_links(links, node_classification, "space_context")
        self.links_sys = _group_links(links, node_classification, "system")

        self.N = len(self.links_sys.keys())
        self.noises = noises
        self.seed = seed
        self.transient_fraction = transient_fraction



    def _constant_over_space(self, data_tc, M):
        data_tc_list = [data_tc for _ in range(M)]
        return data_tc_list

    def _constant_over_time(self, data_sc, T, M, shift):
        data_sc_list = [{i: np.repeat(data_sc[m, i-shift], T) for i in self.links_sc.keys()} for m in
                        range(M)]
        return data_sc_list

    def _generate_temporal_context_data(self, links_tc, T, M, seed):
        """
        Helper Function to generate data for the temporal context nodes. It essentially is a
        wrapper around toys.structural_causal_process to generate data that is random across time
        but constant across datasets.
        """
        if self.noises is not None:
            noises_tc = [self.noises[key] for key in links_tc.keys()]
            if np.all([isinstance(el, np.ndarray) for el in noises_tc]):
                noises_tc = np.stack(noises_tc).transpose()
        else:
            noises_tc = None
        shifted_links_tc = _shift_link_entries(links_tc, -self.N)
        data_tc, nonstat_tc = toys.structural_causal_process(shifted_links_tc, T=T, noises=noises_tc,
                                                             transient_fraction=self.transient_fraction,
                                                             seed=seed)
        data_tc = {i: data_tc[:, i - self.N] for i in links_tc.keys()}

        data_tc_list = self._constant_over_space(data_tc, M)
        return data_tc_list, np.any(nonstat_tc)

    def _generate_spatial_context_data(self, links_sc, T, M, shift, seed):
        """
        Helper Function to generate data for the spatial context nodes. It essentially is a
        wrapper around toys.structural_causal_process to generate data that is random across datasets
        but constant across time.
        """
        shifted_links_sc = _shift_link_entries(links_sc, -shift)
        if self.noises is not None:
            noises_sc = [self.noises[key] for key in links_sc.keys()]
            if np.all([isinstance(el, np.ndarray) for el in noises_sc]):
                noises_sc = np.stack(noises_sc).transpose()
        else:
            noises_sc = None

        data_sc, nonstat_sc = toys.structural_causal_process(shifted_links_sc, T=M, noises=noises_sc,
                                                             transient_fraction=0.,
                                                             seed=seed)

        data_sc_list = self._constant_over_time(data_sc, T, M, shift)
        return data_sc_list, np.any(nonstat_sc)

    def generate_data(self, M, T):
        """
        Generates M datasets of time series generated from a joint structural causal model over different spatial
        and temporal contexts.

         Parameters
         ----------
         M : int
            Number of datasets.
         T : int
            Sample size.

        Returns
        ----------
        data : dictionary with array-like values
            Datasets generated from this process, each dataset has the shape (T, N).
        nonvalid : bool
            Indicates whether data has NaNs or infinities.
        """

        links = {**self.links_tc, **self.links_sc, **self.links_sys}

        K_time = len(self.links_tc.keys())

        data = {}
        nonstationary = []

        time_seed = [1, self.seed]
        space_seed = [2, self.seed]
        system_seed = [3, self.seed]

        # first generate data for temporal context nodes
        data_tc_list, nonstat_tc = self._generate_temporal_context_data(self.links_tc, T, M, time_seed)

        # generate spatial context data (constant in time)
        data_sc_list, nonstat_sc = self._generate_spatial_context_data(self.links_sc,
                                                                       T, M,
                                                                       K_time + self.N,
                                                                       space_seed)
        for m in range(M):
            data_context = dict(data_tc_list[m])
            data_context.update(data_sc_list[m])

            if self.noises is not None:
                noises_filled = self.noises
                if np.all([isinstance(el, np.ndarray) for el in self.noises]):
                    noises_filled = np.copy(self.noises)
                    for key in self.links_sc.keys():
                        # fill up any space-context noise to have T entries, then convert to numpy array
                        noises_filled[key] = np.random.standard_normal(len(self.noises[list(self.links_sys.keys())[0]]))
                    noises_filled = np.stack(noises_filled).transpose()
            else:
                noises_filled = None

            # generate system data that varies over space and time
            data_m, nonstat = toys.structural_causal_process(links, T=T, intervention=data_context,
                                                             transient_fraction=self.transient_fraction,
                                                             seed=system_seed, noises=noises_filled)
            data[m] = data_m
            nonstationary.append(nonstat or nonstat_tc or nonstat_sc)
        return data, np.any(nonstationary)



"""Tigramite causal discovery for time series."""

# Authors: Martin Rabel, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0


import numpy as np
from tigramite.data_processing import DataFrame
from tigramite.causal_effects import CausalEffects


class VariableDescription:
    r"""Variable descrption (base-class)

    Used for mixed variable fitting and mediation, see Mediation-tutorial.

    Parameters
    ----------
    name : string
        The display-name of the variable
    observable : bool
        Is the variable observable or hidden? (used by toy-models)
    exogenous : bool
        Is the variable assumed exogenous (used on noise-terms internally, keep default)
    """

    def __init__(self, name="(unnamed)", observable=True, exogenous=False):
        """Use CategoricalVariable or ContinuousVariable instead."""
        self.noise_term = None
        self.observable = observable
        self.exogenous = exogenous
        self.name = name

    def Noise(self):
        """
        Get the variable-description for the noise-term on this variable.
        Used e.g. for describing SEMs in noise-models

        Returns
        -------
        noise : VariableDescription
            The variable-description for the noise-term on this variable.
        """
        if self.noise_term is None:
            self.noise_term = ExogenousNoiseVariable(self)
        return self.noise_term

    def Lag(self, offset):
        """
        Get the variable-description for the past of this variable.
        Used e.g. for describing SEMs in noise-models

        Parameters
        ----------
        offset : uint
            The lag to introduce (only non-negative values, always in the past).

        Returns
        -------
        lagged-self : VariableDescription
            This variable, at lag "offset"
        """
        if offset == 0:
            return self
        else:
            return LaggedVariable(self, offset)

    def Info(self, detail=0):
        """
        Get an Info-string for this variable.
        (Overridden by derived classes for discrete, continuous, lagged, ...)

        Parameters
        ----------
        detail : uint
            The amount of detail to include.

        Returns
        -------
        info : string
            A description of this variable.
        """
        return self.name

    def PrintInfo(self):
        """Print Info-string (see Info)"""
        print(self.Info())

    def Id(self):
        """Get the 'vanilla' (unlagged etc) description of this variable
        (Overridden by derived classes for discrete, continuous, lagged, ...)

        Returns
        -------
        id : VariableDescription
            A description of the underlying variable of this description.
        """
        return self

    def LagValue(self):
        """Get the time-lag (non-negative, positive values are in the past) of the
        variable described by this indicator.
        (Overridden by derived classes for discrete, continuous, lagged, ...)

        Returns
        -------
        lag : uint
            The time-lag (non-negative, positive values are in the past) of this variable.
        """
        return 0


class LaggedVariable(VariableDescription):
    """Describes a lagged variable (see VariableDescription)"""

    def __init__(self, var, lag):
        """Use var-description.Lag(offset) instead."""
        self.var = var
        self.lag = lag
        self.exogenous = False

    def Id(self):
        return self.var

    def LagValue(self):
        return self.lag

    def Info(self, detail=0):
        return self.var.Info(detail) + " at lag " + str(self.lag)

    def PrintInfo(self):
        print(self.Info())

    def DefaultValue(self):
        return self.var.DefaultValue()


class ExogenousNoiseVariable(VariableDescription):
    """Describes an exogenous noise variable (see VariableDescription)"""

    def __init__(self, attached_to):
        """Use var-description.Noise() instead."""
        super().__init__(name="Noise of " + attached_to.name, observable=False, exogenous=True)
        self.associated_system_variable = attached_to

    def DefaultValue(self):
        return self.associated_system_variable.DefaultValue()


class CategoricalVariable(VariableDescription):
    """Describes a categorical variable (see VariableDescription)

    Parameters
    ----------
    name : string
        The display-name of the variable.
    categories : uint or *iterable*
        The number of categories in which the variable may take values, or
        an iterable containing the possible values.
    dtype : string
        Name of the numpy dtype ('uint32', 'bool', ...) or 'auto'
    """

    def __init__(self, name="(unnamed)", categories=2, dtype="auto"):
        super().__init__(name=name)
        self.is_categorical = True
        if hasattr(categories, '__iter__'):
            self.categories = len(categories)
            self.category_values = categories
        else:
            self.categories = categories
            self.category_values = None
        if dtype == "auto":
            self.dtype = "uint32" if self.categories > 2 else "bool"
        else:
            self.dtype = dtype

    def Info(self, detail=0):
        if detail > 0:
            return self.name
        else:
            return self.name + f"( Categorical variable with {self.categories} categories )"

    def Empty(self, N):
        """Build an empty np.array with N samples."""
        return np.zeros(N, self.dtype)

    def ValidValue(self, x):
        """Validate a value"""
        is_integer = x % 1 == 0.0  # this works for all of: np.[u]int; built-in [u]int; np.bool; built-in boolean
        return is_integer and 0 <= x < self.categories

    def DefaultValue(self):
        """Return a 'default-value' to use in toy-models at the start of a time-series
        (if otherwise undefined by the SEM)"""
        # used for lagged values not available in time-series
        return 0  # could be user specified & per variable ...

    def CopyInfo(self, name="(unnamed)"):
        """Create a new variable-id with the same parameters"""
        return CategoricalVariable(name, self.categories, self.dtype)


class ContinuousVariable(VariableDescription):
    """Describes a continuous variable (see VariableDescription)

       Parameters
       ----------
       name : string
           The display-name of the variable.
       dtype : string
           Name of the numpy dtype ('float32', 'float64', ...)
       dimension : uint
           The number of dimensions, can be >1 for fitting.
       """

    def __init__(self, name="(unnamed)", dtype="float32", dimension=1):
        super().__init__(name=name)
        self.dtype = dtype
        self.is_categorical = False
        self.dimension = dimension  # fitting probabilities for multiple categories might be higher dim (handled by fit)

    def Info(self, detail=0):
        if detail > 0:
            return self.name
        else:
            return self.name + f"( Continuous variable of dimension {self.dimension} )"

    def Empty(self, N):
        """Build an empty np.array with N samples."""
        if self.dimension == 1:
            return np.zeros(N, dtype=self.dtype)
        else:
            return np.zeros([self.dimension, N], dtype=self.dtype)

    def ValidValue(self, x):
        return True

    def DefaultValue(self):
        """Return a 'default-value' to use in toy-models at the start of a time-series
        (if otherwise undefined by the SEM)"""
        # used for lagged values not available in time-series
        return 0.0  # could be user specified & per variable ...

    def CopyInfo(self, name="(unnamed)"):
        """Create a new variable-id with the same parameters"""
        return ContinuousVariable(name, self.dtype, self.dimension)


class Environment:
    """An (exogenous) environment to use with a model

    Counterfactual models may create different worlds with different models from the same environment.

    Parameters
    ----------
    exogenous_noise : dictionary< VariableDescription, *callable* (np.random.generator, sample-count) >
        For each variable, a callable generating noise-values.
    N : uint
        The initial number of samples to generate
    seed : uint or None
        Seed to use for the random-generator.
    """

    def __init__(self, exogenous_noise, N=1000, seed=None):
        self.rng = np.random.default_rng(seed)
        self.can_reset = seed is None
        self.exogenous_noise = exogenous_noise
        self.noise = {}
        self.N = N
        self._ForceReset()

    def ResetWithNewSeed(self, new_seed, N=None):
        """Reset with a new random-seed."""
        self.rng = np.random.default_rng(new_seed)
        self._ForceReset(N)

    def Reset(self, N=None):
        """Reset with a random new random-seed."""
        if not self.can_reset:
            raise Exception("Cannot reset a fixed-seed environment, use seed=None or factory/lambda instead for"
                            "ensemble-creation.")
        self._ForceReset(N)
        return self

    def _ForceReset(self, N=None):
        """[internal] Enfore a reset."""
        if N is not None:
            self.N = N
        for var, _noise in self.exogenous_noise.items():
            self.noise[var.Noise()] = _noise(self.rng, self.N)

    def GetNoise(self):
        """Get a copy of the noise-values.

        Returns
        -------
        noises : dictionary< Variable-Description, samples >
            Returns a shallow copy of the samples generated.
            Ids are 'exogenous_noise.Noise()' (see constructor).
        """
        return self.noise.copy()  # return a SHALLOW copy of the noise-data (copy the dict, not the data)


class DataPointForValidation:
    """[INTERNAL] Used to validate causal ordering on SEM for toy-model"""

    def __init__(self):
        self.known = {}
        self.is_timeseries = False
        self.max_lag = 0

    def Set(self, var, value):
        self.known[var] = True

    def __getitem__(self, key):
        if key.__class__ == LaggedVariable:
            self.is_timeseries = True
            self.max_lag = max(self.max_lag, key.lag)
            return key.DefaultValue()

        if not key.exogenous and key not in self.known:
            raise Exception("SEM must be causally ordered, such that contemporaneous parents are listed "
                            "before their children.")
        return key.DefaultValue()


class DataPointForParents:
    """[INTERNAL] Used to extract causal parents (for ground-truth graph) from SEM for toy-model"""

    def __init__(self):
        self.parents = []

    def __getitem__(self, key):
        if not key.exogenous:
            self.parents.append(key)
        return key.DefaultValue()


class DataPointView:
    """[INTERNAL] Used to run SEMs on toy-models."""

    def __init__(self, data, known=None):
        self.data = data
        self.index = 0

    def Next(self):
        self.index += 1

    def Set(self, var, value):
        self.data[var][self.index] = value

    def __getitem__(self, key):
        if key.__class__ == LaggedVariable:
            if self.index < key.lag:
                return key.DefaultValue()
            else:
                return self.data[key.var][self.index - key.lag]
        else:
            return self.data[key][self.index]


class Model:
    """Describes a Toy-model via an SEM.

    Validates only the causal ordering, for more validation, see tigramite.toymodels.
    However, this implementation can also generate non-additive models.

    Parameters
    ----------
    sem : dictionary< VariableDescription, *callable* ( sample-view : var-desc->np.array or scalar )
        For variables in causal order, a *callable* which is passed a
        view v of the data-samples, access data by v[var-description],
        and return the value(s) of the described (key) variable.
        For non-time-series, v[var-description] is an np.array,
        for time-series, v[var-description] is a scalar.
    """

    def __init__(self, sem):
        self.SEM = sem
        self.is_timeseries = False
        self.max_lag = 0
        self.Validate()

    def GetGroundtruthLinks(self):
        """Get the causal links to parents (ground-truth).

        Does not validate faithfullness.

        Returns
        -------
        links, indices : dictionaries< VariableDescription, ...>
            For links: Values are the lists of parents.
            For indices: Values are variable indices (see GetGroundtruthLinksRaw)
        """
        self.Validate()
        links = {}
        indices = {}
        idx = 0
        for var, eq in self.SEM.items():
            data_pt = DataPointForParents()
            self.SEM[var](data_pt)
            links[var] = data_pt.parents
            indices[var] = idx
            idx += 1
        return links, indices

    def GetGroundtruthLinksRaw(self):
        """Get unformatted causal links to parents (ground-truth).

        Does not validate faithfullness.

        Returns
        -------
        links : tigramite-format/raw indices + lags
        """
        links, indices = self.GetGroundtruthLinks()
        links_raw = {}
        for var in self.SEM.keys():
            parents_raw = []
            for p in links[var]:
                if p.__class__ == LaggedVariable:
                    parents_raw.append((indices[p.var], -p.lag))
                else:
                    parents_raw.append((indices[p], 0))
            links_raw[indices[var]] = parents_raw
        return links_raw

    def GetGroundtruthGraph(self):
        """Get the (ground-truth) graph.

        Returns
        -------
        graph, graph-type : see e.g. CausalEffects tutorial, string
        """
        graph = CausalEffects.get_graph_from_dict(self.GetGroundtruthLinksRaw())
        if self.is_timeseries:
            return graph, 'stationary_dag'
        else:
            return graph, 'dag'

    def Validate(self):
        """Validate causal ordering (necessary for consistent data-generation), called automatically."""
        data_pt = DataPointForValidation()
        for var, eq in self.SEM.items():
            data_pt.Set(var, self.SEM[var](data_pt))
        self.max_lag = max(self.max_lag, data_pt.max_lag)
        self.is_timeseries = data_pt.is_timeseries

    def ApplyWithExogenousNoise(self, environment, partial_data=None):
        """Apply to environment

        Parameters
        ----------
        environment : Environment
            Exogenous noise-samples given as environment.
        partial_data : None
            [INTERNAL USE] Leave to default.
        """
        if self.is_timeseries:
            # This may be very slow (as it cannot be parallelized or be dispatched efficiently to native code via numpy)
            return self._GenerateAsTimeseries(environment, partial_data)

        data = environment.GetNoise()
        vars = []
        for var, eq in self.SEM.items():
            if partial_data is not None and var in partial_data:
                data[var] = partial_data[var]
            else:
                vars.append(var)

        for var in vars:
            data[var] = self.SEM[var](data)
        return data

    def _GenerateAsTimeseries(self, environment, partial_data=None):
        """[INTERNAL] Generate time-series data.

        Generation of time-series data cannot be parallelized of efficiently dispatched to native code.
        """
        data = environment.GetNoise()
        vars = []
        for var, eq in self.SEM.items():
            if partial_data is not None and var in partial_data:
                data[var] = partial_data[var]
            else:
                data[var] = var.Empty(environment.N)
                vars.append(var)

        data_pt = DataPointView(data, partial_data)
        for x in range(environment.N):
            for var in vars:
                data_pt.Set(var, self.SEM[var](data_pt))
            data_pt.Next()
        return data

    def Intervene(self, changes):
        """Get and intervened model.

        Parameters
        ----------
        changes : dictionary< VariableDescription, ...>
            For each variable to intervene on, either a scalar (hard intervention),
            or a *callable* replacing the equation in the SEM (see constructor).

        Returns
        -------
        intervened model : Model
            The intervened model.
        """
        # Return a model, that describes the intervened system

        new_sem = self.SEM.copy()
        for var, eq in changes.items():
            if callable(eq):
                new_sem[var] = eq  # intervene by function
            else:
                new_sem[var] = lambda data: eq  # return a constant
        return Model(new_sem)


class World:
    """A 'world' instantiation.

    Generate observations from an environment (exogenous noise) and a model (SEM).

    Parameters
    ----------
    environment : Environment
        The environment with exogenous noise samples.
    model : Model
        The SEM describing the system.
    """

    def __init__(self, environment, model):
        self.environment = environment  # to check agreement in counterfactual worlds
        if model is not None:
            self.data = model.ApplyWithExogenousNoise(environment)
        else:
            self.data = {}

    def Observables(self):
        """Get all observables

        Returns
        -------
        observables : dictonary< VariableDescription, np.array( environment.N ) >
            The samples for each observable variable.
        """
        obs = {}
        for var, values in self.data.items():
            if var.observable:
                obs[var] = values
        return obs


class CounterfactualWorld(World):
    """A 'counterfactual' world.

    Generate observations from one environment (exogenous noise) and multiple models (SEM).

    Parameters
    ----------
    environment : Environment
        The environment with exogenous noise samples.
    base-model : Model
        The base-model to generate data from in the end.
        Call TakeVariablesFromWorld overwrite some variables with values from another 'world',
        then call Compute.
    """

    def __init__(self, environment, model):
        super().__init__(environment, None)
        self.model = model

    def TakeVariablesFromWorld(self, world, vars):
        """Take variables from a world.

        Parameters
        ----------
        world : World
            Take samples from here.
        vars : VariableDescription or *iterable* <VariableDescription>
            One (or a iterable of many) variable to overwrite.
        """
        if world.environment != self.environment:
            raise Exception("Counterfactual Worlds must share exogenous noise terms.")

        if hasattr(vars, '__iter__'):
            for var in vars:
                self.TakeVariablesFromWorld(world, var)
        else:
            self.data[vars] = world.data[vars]

    def Compute(self):
        """Compute the counterfactual world observations."""
        self.data = self.model.ApplyWithExogenousNoise(self.environment, partial_data=self.data)


class DataPointView_TS_Window:
    """[INTERNAL] Injected into SEM callbacks to compute values for counterfactual windows in parallel."""

    def __init__(self, env, stationary_data, window_data, offset_in_window, max_lag):
        self.stationary_data = stationary_data
        self.environment = env
        self.window_data = window_data
        self.offset_in_window = offset_in_window
        self.max_lag = max_lag

    def __getitem__(self, key):
        if key.exogenous:
            return self.environment.noise[key][self.max_lag:]
        else:
            if self.offset_in_window < key.LagValue():
                return self.stationary_data[key.Id()][self.max_lag - key.LagValue():-key.LagValue()]
            else:
                return self.window_data[self.offset_in_window - key.LagValue()][key.Id()]


class CounterfactualTimeseries:
    """[INTERNAL] Used to generate ground-truth for time-series NDE.

    Computes the effect of an intervention at a SINGLE point in time.
    Use GroundTruth_* functions instead.
    """

    def __init__(self, environment, base_model, max_lag_in_interventions):
        self.environment = environment
        self.base_model = base_model
        self.max_lag = max_lag_in_interventions

        self.stationary_data = self.base_model.ApplyWithExogenousNoise(self.environment)

    def ComputeIntervention(self, output_var, interventions, take):
        output_windows = []
        for var, value in interventions:
            intervened_window = []
            output_windows.append(intervened_window)
            found_in_window = False
            for delta_t in range(self.max_lag + 1):
                data_pts = DataPointView_TS_Window(self.environment, self.stationary_data,
                                                   intervened_window, delta_t, self.max_lag)
                current_time = {}
                intervened_window.append(current_time)
                for v, eq in self.base_model.SEM.items():
                    if v.Id() == var.Id() and self.max_lag - delta_t == var.LagValue():
                        found_in_window = True
                        current_time[v] = value  # this is a (single) value, but numpy can broadcast it
                    else:
                        current_time[v] = eq(data_pts)
            if not found_in_window:
                raise Exception(f"Intervention {var.Info()}={value} was not found in time-series window.")
        # default to window 0
        cf_window = output_windows[0]
        for m in take:
            cf_window[self.max_lag - m.LagValue()][m.Id()] = output_windows[1][self.max_lag - m.LagValue()][m.Id()]
        data_pts = DataPointView_TS_Window(self.environment, self.stationary_data,
                                           cf_window, self.max_lag, self.max_lag)
        return self.base_model.SEM[output_var](data_pts)


def Ensemble(shared_setup, payloads, runs=1000):
    """Helper to run e.g. estimator vs ground-truth on an ensemble of model-realizations"""

    results = np.zeros([len(payloads), runs])
    get_next = shared_setup
    if not callable(shared_setup):
        environment = shared_setup
        get_next = lambda: environment.Reset()

    for r in range(runs):
        environment = get_next()
        p_idx = 0
        for p in payloads:
            results[p_idx, r] = p(environment)
            p_idx += 1
    return results


def _Fct_on_grid(fct, list_of_points, cf_delta=0.5, normalize_by_delta=False, **kwargs):
    """Helper to evaluate 'fct' on a grid of points"""

    result = []
    for pt in list_of_points:
        result.append(fct(pt, pt + cf_delta, **kwargs))

    if normalize_by_delta:
        return np.array(result) / cf_delta
    else:
        return np.array(result)


def _Fct_smoothed(fct, min_x, max_x, cf_delta=0.5, steps=100, smoothing_gaussian_sigma_in_steps=5,
                  normalize_by_delta=False, boundary_effects="extend range", **kwargs):
    """Helper to evaluate 'fct' on a grid of points with subsequent Gauß-smoothing"""

    stepsize = (max_x - min_x) / steps
    # Extend the window to run on (numpy would extend by zeros if mode="same")
    if boundary_effects == "extend range":
        steps += 6 * smoothing_gaussian_sigma_in_steps - 1
        min_x -= 3 * smoothing_gaussian_sigma_in_steps * stepsize
    else:
        raise Exception("Currently only smoothing mode for boundary-effects is 'extend range'")

    x_values = []  # np.arange with floats is unstable wrt len of output
    result = []
    for i in range(steps + 1):
        pt = stepsize * i + min_x
        x_values.append(pt)
        value = fct(pt, pt + cf_delta, **kwargs)
        result.append(value)

    if normalize_by_delta:
        result = np.array(result) / cf_delta
    else:
        result = np.array(result)

    # cut off convolution-kernel at 3 sigma
    gx = np.arange(-3 * smoothing_gaussian_sigma_in_steps, 3 * smoothing_gaussian_sigma_in_steps)
    gaussian = (np.exp(-(gx / smoothing_gaussian_sigma_in_steps) ** 2 / 2)
                / np.sqrt(2.0 * np.pi) / smoothing_gaussian_sigma_in_steps)
    if len(np.shape(result)) == 1:  # values = (samples)
        smoothed = np.convolve(result, gaussian, mode="valid")
    elif len(np.shape(result)) == 3:  # densities = (samples, categories, cf/te)
        smoothed = np.empty_like(
            result[3 * smoothing_gaussian_sigma_in_steps:1 - 3 * smoothing_gaussian_sigma_in_steps, :, :])
        for cf_te in range(2):
            for p_category in range(np.shape(result)[1]):
                smoothed[:, p_category, cf_te] = np.convolve(result[:, p_category, cf_te], gaussian, mode="valid")
    else:
        raise Exception(f"Invalid result-shape {result.shape}")

    return np.array(x_values)[3 * smoothing_gaussian_sigma_in_steps:1 - 3 * smoothing_gaussian_sigma_in_steps], smoothed


def PlotInfo(**va_arg_dict):
    """Helper to add info to plot-setup"""
    return va_arg_dict


def PlotAbsProbabilities(plt, target_var, data, labels):
    """Helper to plot Effects on Categorical variables (typically set plt=your-pylot-module-name)"""

    fig = plt.figure(figsize=[12.0, 4.8], dpi=75.0, layout='constrained')
    fig.suptitle(labels["title"])
    shared_ax = None
    has_printed_legend = False
    for cY in range(target_var.categories):
        if shared_ax is None:
            shared_ax = plt.subplot(131 + cY)  # digits are rows, cols, index+1
        else:
            plt.subplot(131 + cY, sharey=shared_ax)  # digits are rows, cols, index+1
        for d in data:
            plt.plot(d["x"], d["y"][:, cY, 1], color=d["colorTE"], label=d["labelTE"])
            plt.plot(d["x"], d["y"][:, cY, 0], color=d["colorCF"], label=d["labelCF"])
        plt.xlabel(labels["x"])
        plt.ylabel(labels["y"].format(cY=cY))
        if not has_printed_legend:
            has_printed_legend = True
            fig.legend()
    return fig


def PlotChangeInProbabilities(plt, target_var, data, labels):
    """Helper to plot Effects on Categorical variables (typically set plt=your-pylot-module-name)"""

    fig = plt.figure(figsize=[12.0, 4.8], dpi=75.0, layout='constrained')
    fig.suptitle(labels["title"])
    shared_ax = None
    has_printed_legend = False
    for cY in range(target_var.categories):
        if shared_ax is None:
            shared_ax = plt.subplot(131 + cY)  # digits are rows, cols, index+1
        else:
            plt.subplot(131 + cY, sharey=shared_ax)  # digits are rows, cols, index+1
        for d in data:
            plt.plot(d["x"], d["y"][:, cY, 0] - d["y"][:, cY, 1], color=d["color"], label=d["label"])
        plt.xlabel(labels["x"])
        plt.ylabel(labels["y"].format(cY=cY))
        if not has_printed_legend:
            has_printed_legend = True
            fig.legend()
    return fig


def FindMaxLag(*va_args):
    """[INTERNAL] Finds the max lag in a collection of variables."""
    max_lag = 0
    for var_group in va_args:
        if hasattr(var_group, '__iter__'):
            for var in var_group:
                max_lag = max(max_lag, var.LagValue())
        else:
            var = var_group
            max_lag = max(max_lag, var.LagValue())
    return max_lag


def GroundTruth_NDE_auto(change_from, change_to, estimator, env, model):
    """Ground-Truth computation from toy-model for NDE.

    GroundTruth_*_auto functions extract source, target, blocked-mediators from an estimator.

    Parameters
    ----------
    change_from : single value of same type as single sample for X (float, int, or bool)
        Reference-value to which X is set by intervention in the world seen by the mediator.
    change_to : single value of same type as single sample for X (float, int, or bool)
        Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).
    estimator : NaturalEffects_GraphMediation
        Extract source, target, blocked-mediators from estimator.
    env : Environment
        The environment used.
    model : Model
        The toy-model.

    Returns
    -------
    NDE : If Y is categorical -> np.array( # categories Y, 2 )
        The probabilities the categories of Y (after, before) changing the interventional value of X
        as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.

    NDE : If Y is continuous -> float
        The change in the expectation-value of Y induced by changing the interventional value of X
        as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.
    """
    return GroundTruth_NDE(change_from, change_to, estimator.Source, estimator.Target, estimator.BlockedMediators,
                           env, model)


def GroundTruth_NDE(change_from, change_to, source, target, mediators, env, model):
    """Ground-Truth computation from toy-model for NDE.

    Note: GroundTruth_*_auto functions extract source, target, blocked-mediators from an estimator.

    Parameters
    ----------
    change_from : single value of same type as single sample for X (float, int, or bool)
        Reference-value to which X is set by intervention in the world seen by the mediator.
    change_to : single value of same type as single sample for X (float, int, or bool)
        Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).
    source : VariableDescription
        Effect source.
    target : VariableDescription
        Effect target.
    mediators : *iterable* <VariableDesciption>
        Blocked mediators.
    env : Environment
        The environment used.
    model : Model
        The toy-model.

    Returns
    -------
    NDE : If Y is categorical -> np.array( # categories Y, 2 )
        The probabilities the categories of Y (after, before) changing the interventional value of X
        as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.

    NDE : If Y is continuous -> float
        The change in the expectation-value of Y induced by changing the interventional value of X
        as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.
    """

    if target.LagValue() != 0:
        raise Exception("Do not use lagged targets, it is always possible to shift everything, so that the"
                        "effect is on an unlagged variable.")

    if model.is_timeseries:
        # Generate groundtruth for intervention at a SINGLE point in time

        # Window-size (at each point) must be at least the
        max_lag_in_interventions = FindMaxLag(source, target, mediators)
        # shouldn't be necessary, but fixes some issues with offsets in stationary data:
        max_lag_in_interventions = max(max_lag_in_interventions, model.max_lag)

        ts = CounterfactualTimeseries(env, model, max_lag_in_interventions=max_lag_in_interventions)

        y_cf = ts.ComputeIntervention(target, [(source, change_to), (source, change_from)], mediators)
        y_real = ts.ComputeIntervention(target, [(source, change_from)], [])

    else:
        # Ground-Truth for non-timeseries is straight-forward:

        modelA = model.Intervene(changes={source.Id(): change_from})
        modelB = model.Intervene(changes={source.Id(): change_to})
        worldA = World(env, modelA)
        worldB = World(env, modelB)
        cf_world = CounterfactualWorld(env, model)
        cf_world.TakeVariablesFromWorld(worldA, mediators)
        cf_world.TakeVariablesFromWorld(worldB, source)
        cf_world.Compute()

        y_cf = cf_world.Observables()[target]
        y_real = worldA.Observables()[target]

    if target.is_categorical:
        result = []
        for cY in range(target.categories):
            result.append([np.count_nonzero(y_cf == cY), np.count_nonzero(y_real == cY)])
        return np.array(result) / env.N
    else:
        return np.mean(y_cf - y_real)


def GroundTruth_NDE_fct_auto(x_min, x_max, estimator, env, model, cf_delta=0.5, normalize_by_delta=False,
                             grid_stepping=0.1):
    """See GroundTruth_NDE_auto and NaturalEffects_GraphMediation.NDE_smoothed."""
    source = estimator.Source
    target = estimator.Target
    blocked_mediators = estimator.BlockedMediators
    grid = np.arange(x_min, x_max, grid_stepping)
    return grid, GroundTruth_NDE_fct(source, target, blocked_mediators, env, model, grid, cf_delta, normalize_by_delta)


def GroundTruth_NDE_fct(source, target, mediators, env, model, list_of_points, cf_delta=0.5, normalize_by_delta=False):
    """See GroundTruth_NDE and NaturalEffects_GraphMediation.NDE_smoothed."""
    return _Fct_on_grid(GroundTruth_NDE, list_of_points, cf_delta, normalize_by_delta,
                        source=source, target=target, mediators=mediators, env=env, model=model)


def GroundTruth_NIE_fct(source, target, mediator, env, model, list_of_points, cf_delta=0.5, normalize_by_delta=False):
    """See GroundTruth_NIE and NaturalEffects_StandardMediation.NIE_smoothed."""
    return _Fct_on_grid(GroundTruth_NIE, list_of_points, cf_delta, normalize_by_delta,
                        source=source, target=target, mediator=mediator, env=env, model=model)


def GroundTruth_NIE(change_from, change_to, source, target, mediator, env, model):
    """Ground-Truth computation from toy-model for NIE.

    Standard-mediation setup only.

    Parameters
    ----------
    change_from : single value of same type as single sample for X (float, int, or bool)
        Reference-value to which X is set by intervention in the world seen by the mediator.
    change_to : single value of same type as single sample for X (float, int, or bool)
        Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).
    source : VariableDescription
        Effect source.
    target : VariableDescription
        Effect target.
    mediator : VariableDesciption
        Effect mediator.
    env : Environment
        The environment used.
    model : Model
        The toy-model.

    Returns
    -------
    NIE : If Y is categorical -> np.array( # categories Y, 2 )
        The probabilities the categories of Y (after, before) changing the interventional value of X
        as "seen" by M from change_from to change_to, while keeping the value as (directly) seen by Y,
        as if X remained at change_from.

    NIE : If Y is continuous -> float
        The change in the expectation-value of Y induced by changing the interventional value of X
        as "seen" by M from change_from to change_to, while keeping the value as (directly) seen by Y,
        as if X remained at change_from.
    """

    modelA = model.Intervene(changes={source: change_from})
    modelB = model.Intervene(changes={source: change_to})
    worldA = World(env, modelA)
    worldB = World(env, modelB)
    cf_world = CounterfactualWorld(env, model)
    # this time, take the mediator from the B-world
    cf_world.TakeVariablesFromWorld(worldA, [source])
    cf_world.TakeVariablesFromWorld(worldB, [mediator])
    cf_world.Compute()

    y_cf = cf_world.Observables()[target]
    y_real = worldA.Observables()[target]

    if target.is_categorical:
        result = []
        for cY in range(target.categories):
            result.append([np.count_nonzero(y_cf == cY), np.count_nonzero(y_real == cY)])
        return np.array(result) / env.N
    else:
        return np.mean(y_cf - y_real)


class DataHandler:
    """[INTERNAL] Implement some helper functions and generate time-series data via tigramite's data-frames."""

    def __init__(self, observables, dataframe_based_preprocessing=True):
        if isinstance(observables, DataFrame):
            self._from_dataframe = True
            self._data = VariablesFromDataframe(observables)
            self._dataframe = observables
            self.use_dataframe_for_preprocessing = dataframe_based_preprocessing
            self.data_selection_virtual_ids = {}
            self._info = {}
            self.preprocessed_data = False
            self.data_selection_frozen = False # filter all data for all fits for missing values, lock after first data-access (see GetVariableAuto and operator [] / __getitem__)
        else:
            self._from_dataframe = False
            self._data = observables
            self._dataframe = DataframeFromVariables(observables)            
            self.use_dataframe_for_preprocessing = False

        self._indices = {}
        self._keys = list(self._data.keys())
        idx = 0
        for var in self._data.keys():
            self._indices[var] = idx
            idx += 1


    def GetIdFor(self, idx, lag, display_name):
        # self._keys[idx] contains info (eg catgorical or continuous, 'original name') for variable idx, attach display name (eg Mediator)
        # also add 'original name' eg 'temperature' in square brackets as well as lag
        base_id =  self._keys[idx]
        return base_id.CopyInfo(display_name + f"[{base_id.Info(detail=1)} at lag {lag}]")
    
    def RequirePreprocessingFor(self, var, info):
        if self.use_dataframe_for_preprocessing:            
            if var not in self.data_selection_virtual_ids:
                # if variable locked also ok after frozen
                if self.data_selection_frozen:
                    raise Exception("Cannot add variables to preprocessing after variable-set has been locked-in")
                assert info is not None
                var_id = self.GetIdFor(var[0], var[1], info)
                self.data_selection_virtual_ids[var] = var_id
                self._indices[var_id] = var
                self._info[var_id] = info

    def GetVariableAuto(self, var, info="Other"):
        if self._from_dataframe:
            return self.ReverseLookupSingle(var, info)
        else:
            return var

    def GetVariablesAuto(self, vars, info="Other"):
        result = []
        for var in vars:
            result.append(self.GetVariableAuto(var, info))
        return result

    def DataFrame(self):
        return self._dataframe

    def __getitem__(self, key):
        if hasattr(key, '__iter__'):
            return [self[entry] for entry in key]
        
        if self.use_dataframe_for_preprocessing:
            return self._indices[key]

        if key.__class__ == LaggedVariable:
            return self._indices[key.var], -key.lag
        else:
            return self._indices[key], 0

    def _PreprocessData(self, **kwargs):
        self.data_selection_frozen = True
        self.preprocessed_data = True
        X = []
        Y = []
        Z = []
        M = []
        for var_lag_index, var_id in self.data_selection_virtual_ids.items():
            info = self._info[var_id]
            if info == "Source":
                X.append(var_lag_index)
            elif info == "Target":
                Y.append(var_lag_index)
            elif info == "Mediator":
                M.append(var_lag_index)
            elif info == "Adjustment":
                Z.append(var_lag_index)
            else:
                raise Exception("Unknown Variable-Interpretation")
        data_preprocessed, xyz, data_type = self.DataFrame().construct_array(X=X, Y=Y, Z=Z, extraZ=M, **kwargs) # kw-args forwards eg tau-max
        self._data = {}
        i = 0
        for x in X:
            self._data[self.data_selection_virtual_ids[x]] = data_preprocessed[i]
            assert xyz[i] == 0
            i += 1
        for y in Y:
            self._data[self.data_selection_virtual_ids[y]] = data_preprocessed[i]
            assert xyz[i] == 1
            i += 1
        for z in Z:
            self._data[self.data_selection_virtual_ids[z]] = data_preprocessed[i]
            assert xyz[i] == 2
            i += 1
        for m in M:
            self._data[self.data_selection_virtual_ids[m]] = data_preprocessed[i]
            assert xyz[i] == 3
            i += 1
        assert i == data_preprocessed.shape[0]
            
    def GetPreprocessed(self, vars, **kwargs):
        if not self.preprocessed_data:
            self._PreprocessData(**kwargs)
        result = {}
        for var in vars:
            assert var in self.data_selection_virtual_ids
            var_id = self.data_selection_virtual_ids[var]
            result[var_id] = self._data[var_id]
        return list(result.keys()), result
            

    def Get(self, name, vars, **kwargs):
        if self.use_dataframe_for_preprocessing:
            return self.GetPreprocessed(vars, **kwargs)
        ids = []
        i = 1
        for idx, lag in vars:
            # assemble names for this variable-group, eg "Mediator5"
            the_name = name if len(vars) == 1 else name + str(i)
            ids.append( self.GetIdFor(idx, lag, the_name) )
            i += 1
        data, xyz, data_type = self.DataFrame().construct_array(X=vars, Y=[], Z=[], **kwargs)
        assert data.shape[0] == len(ids)
        result = {}
        i = 0
        for elem in ids:
            result[elem] = data[i].astype(dtype=np.dtype(elem.dtype))
            i += 1
        return ids, result

    def ReverseLookupSingle(self, index, info):
        if self.use_dataframe_for_preprocessing:
            self.RequirePreprocessingFor(index, info)
            return self.data_selection_virtual_ids[index]
        else:
            return self._keys[index[0]].Lag(-index[1])

    def ReverseLookupMulti(self, index_set, info=None):
        return [self.ReverseLookupSingle(index, info) for index in index_set]


def VariablesFromDataframe(dataframe):
    """Extract Category-Information from tigramite::dataframe

    Parameters
    ----------
    dataframe : tigramite.data_processing.DataFrame
        Dataframe to extract data from.


    Returns
    -------
    variables : dictionary< VariableDescription, np.array(N) >
        The variable-meta-data and data.
    """

    data_types = dataframe.data_type
    if dataframe.data_type is not None:
        # Require data-types to be constant
        first_elements = data_types[0, :]
        if not np.all(first_elements == data_types):
            raise NotImplementedError("Natural Effect Framework currently only supports per variable"
                                      "data-types, this dataframe contains variables with changing"
                                      "(over time) data-types.")

    result = {}  # var i will be in result.items()[i]
    data = dataframe.values[0]
    node_count = np.shape(data)[1]
    for node in range(node_count):
        if data_types is None or data_types[0, node] == 0:
            var = ContinuousVariable(name=str(dataframe.var_names[node]))
            result[var] = data[:, node]
        else:
            labels, transformed = np.unique(data[:, node], return_inverse=True)
            var = CategoricalVariable(name=str(dataframe.var_names[node]), categories=labels)
            result[var] = transformed
    return result


def DataframeFromVariables(data_dict):
    """Convert Category-Information to tigramite::dataframe

    Parameters
    ----------
    data_dict : dictionary< VariableDescription, np.array(N) >
        The variable-meta-data and data.

    Returns
    -------
    dataframe : tigramite.data_processing.DataFrame
        Dataframe containing the raw data.
    """

    data_len = 0
    keys = list(data_dict.keys())
    values = list(data_dict.values())
    if keys[0].is_categorical or keys[0].dimension == 1:
        data_len = np.shape(values[0])[0]
    else:
        data_len = np.shape(values[0])[1]
    dimensions = 0
    for var, data in data_dict.items():
        if var.is_categorical:
            dimensions += 1
        else:
            dimensions += var.dimension

    data_out = np.zeros([data_len, dimensions])
    data_type = np.zeros([data_len, dimensions])
    names = []
    i = 0
    for var, data in data_dict.items():
        if var.is_categorical:
            names.append(var.name)
            if var.category_values is not None:
                data_out[:, i] = np.take(var.category_values, data)
            else:
                data_out[:, i] = data
            data_type[:, i] = np.ones(data_len)
            i += 1
        else:
            if var.dimension > 1:
                for j in range(var.dimension):
                    if var.dimension == 1:
                        names.append(var.name)
                    else:
                        names.append(var.name + "$_" + str(j) + "$")
                    data_out[:, i + j] = data[j]
                    data_type[:, i + j] = np.zeros(data_len)
            else:
                names.append(var.name)
                data_out[:, i] = data
                data_type[:, i] = np.zeros(data_len)
            i += var.dimension
    return DataFrame(data=data_out, data_type=data_type, var_names=names)



"""
Tests for pcmci.py, including tests for run_pc_stable, run_mci, and run_pcmci.
"""
from __future__ import print_function
import numpy as np
import pytest
import sklearn.linear_model

from tigramite.models import LinearMediation
import tigramite.data_processing as pp
from tigramite.toymodels import structural_causal_processes as toys
from tigramite.models import Prediction
from tigramite.independence_tests.parcorr import ParCorr

from test_pcmci_calculations import a_chain, gen_data_frame

# Pylint settings
# pylint: disable=redefined-outer-name

# TEST LINK GENERATION #########################################################
@pytest.fixture(params=[
    # Generate a test data sample
    # Parameterize the sample by setting the autocorrelation value, coefficient
    # value, total time length, and random seed to different numbers
    # links_coeffs,               time, seed_val
    (a_chain(0.9, 0.7), 10000, 21),
    (a_chain(0.9, 0.7), 10000, 42)])
def data_frame_a(request):
    # Set the parameters
    links_coeffs, time, seed_val = request.param
    # Generate the dataframe
    return gen_data_frame(links_coeffs, time, seed_val), links_coeffs

# TEST MODELS ##################################################################
def test_linear_mediation_coeffs(data_frame_a):
    # Build the dataframe and the model
    (dataframe, true_parents), links_coeffs = data_frame_a
    med = LinearMediation(dataframe=dataframe, data_transform=None)
    # Fit the model
    med.fit_model(all_parents=true_parents, tau_max=3)
    # Ensure the results make sense
    for j, i, tau, coeff in toys._iter_coeffs(links_coeffs):
        np.testing.assert_allclose(med.get_coeff(i=i, tau=tau, j=j),
                                   coeff, rtol=1e-1)

def test_linear_med_cause_coeffs(data_frame_a):
    # Build the dataframe and the model
    (dataframe, true_parents), _ = data_frame_a
    med = LinearMediation(dataframe=dataframe, data_transform=None)
    # Fit the model
    med.fit_model(all_parents=true_parents, tau_max=3)
    # Ensure the causal and mediated causal effects make sense
    for causal_coeff in [med.get_ce(i=0, tau=-2, j=2),
                         med.get_mce(i=0, tau=-2, j=2, k=1)]:
        np.testing.assert_allclose(causal_coeff, 0.49, rtol=1e-1)

# TEST PREDICTIONS #############################################################
def test_predictions(data_frame_a):
    # TODO NOTE: This doesn't actually test if the predictions make sense, only 
    # that they work!
    # Get the data
    (dataframe, true_parents), links_coeffs = data_frame_a
    T = dataframe.T[0]
    # Build the prediction
    a_cond_ind_test = ParCorr(significance='analytic')
    pred = Prediction(dataframe=dataframe,
                      cond_ind_test=a_cond_ind_test,
                      prediction_model=sklearn.linear_model.LinearRegression(),
                      train_indices=range(int(0.8*T)),
                      test_indices=range(int(0.8*T), T),
                      verbosity=0)
    # Load some parameters
    tau_max = 3
    steps_ahead = 1
    target = 2
    # Get the predictors from pc_stable
    all_predictors = pred.get_predictors(selected_targets=[target],
                                         selected_links=None,
                                         steps_ahead=steps_ahead,
                                         tau_max=tau_max,
                                         pc_alpha=None,
                                         max_conds_dim=None,
                                         max_combinations=1)
    # Fit the predictors using the ML method
    pred.fit(target_predictors=all_predictors,
             selected_targets=[target],
             tau_max=tau_max,
             return_data=True)
    # Predict the values
    _ = pred.predict(target)
    # TODO test the values make sense



"""Tigramite data processing functions."""

# Authors: Jakob Runge <jakob@jakob-runge.com>
#          Andreas Gerhardus <andreas.gerhardus@dlr.de>
# License: GNU General Public License v3.0

from __future__ import print_function
from collections import defaultdict, OrderedDict
import sys
import warnings
from copy import deepcopy
import math
import numpy as np
import scipy.sparse
import scipy.sparse.linalg
from scipy import stats
# from numba import jit

class DataFrame():
    """Data object containing single or multiple time series arrays and optional 
    mask, as well as variable definitions.

    Parameters
    ----------
    data : array-like
        if analysis_mode == 'single':
         Numpy array of shape (observations T, variables N)
         OR
         Dictionary with a single entry whose value is a numpy array of
         shape (observations T, variables N)
        if analysis_mode == 'multiple':
         Numpy array of shape (multiple datasets M, observations T,
         variables N)
         OR
         Dictionary whose values are numpy arrays of shape
         (observations T_i, variables N), where the number of observations
         T_i may vary across the multiple datasets but the number of variables
         N is fixed. 
    mask : array-like, optional (default: None)
        Optional mask array, must be of same format and shape as data.
    data_type : array-like
        Binary data array of same shape as array which describes whether 
        individual samples in a variable (or all samples) are continuous 
        or discrete: 0s for continuous variables and 1s for discrete variables.
    missing_flag : number, optional (default: None)
        Flag for missing values in dataframe. Dismisses all time slices of
        samples where missing values occur in any variable. For
        remove_missing_upto_maxlag=True also flags samples for all lags up to
        2*tau_max (more precisely, this depends on the cut_off argument in
        self.construct_array(), see further below). This avoids biases, see
        section on masking in Supplement of Runge et al. SciAdv (2019).
    vector_vars : dict
        Dictionary of vector variables of the form,
        Eg. {0: [(0, 0), (1, 0)], 1: [(2, 0)], 2: [(3, 0)], 3: [(4, 0)]}
        The keys are the new vectorized variables and respective tuple values
        are the individual components of the vector variables. In the method of
        construct_array(), the individual components are parsed from vector_vars
        and added (accounting for lags) to the list that creates X, Y and Z for
        conditional independence test.
    var_names : list of strings, optional (default: range(N))
        Names of variables, must match the number of variables. If None is
        passed, variables are enumerated as [0, 1, ...]
    datatime : array-like, optional (default: None)
        Timelabel array. If None, range(T) is used.
    remove_missing_upto_maxlag : bool, optional (default: False)
        Whether to remove not only missing samples, but also all neighboring
        samples up to max_lag (as given by cut_off in construct_array).
    analysis_mode : string, optional (default: 'single')
        Must be 'single' or 'multiple'.
        Determines whether data contains a single (potentially multivariate)
        time series (--> 'single') or multiple time series (--> 'multiple').
    reference_points : None, int, or list (or 1D array) of integers,
        optional (default:None)
        Determines the time steps --- relative to the shared time axis as
        defined by the optional time_offset argument (see below) --- that are
        used to create samples. Set to [0, 1, ..., T_max-1] if None is passed, 
        where T_max is self.largest_time_step, see below.
        All values smaller than 0 and bigger than T_max-1 will be ignored.
        At least one value must be in [0, 1, ..., T_max-1].
    time_offsets : None or dict, optional (default: None)
        if analysis_mode == 'single':
         Must be None.
         Shared time axis defined by the time indices of the single time series
        if analysis_mode == 'multiple' and data is numpy array:
         Must be None.
         All datasets are assumed to be already aligned in time with
         respect to a shared time axis, which is the time axis of data
        if analysis_mode == 'multiple' and data is dictionary:
         Must be dictionary of the form {key(m): time_offset(m), ...} whose
         set of keys agrees with the set of keys of data and whose values are
         non-negative integers, at least one of which is 0. The value
         time_offset(m) defines the time offset of dataset m with
         respect to a shared time axis.

    Attributes
    ----------
    self._initialized_from : string
        Specifies the data format in which data was given at instantiation.
        Possible values: '2d numpy array', '3d numpy array', 'dict'.
    self.values : dictionary
        Dictionary holding the observations given by data internally mapped to a
        dictionary representation as follows:
        If analysis_mode == 'single': for self._initialized_from == '2d numpy array' this
        is {0: data} and for self._initialized_from == 'dict' this is data.
        If analysis_mode == 'multiple': If self._initialized_from == '3d numpy array', this is
        {m: data[m, :, :] for m in range(data.shape[0])} and for self._initialized_from == 'dict' this
        is data.
    self.datasets: list
        List of the keys identifiying the multiple datasets, i.e.,
        list(self.values.keys())
    self.mask : dictionary
        Mask internally mapped to a dictionary representation in the same way as
        data is mapped to self.values
    self.data_type : array-like
        Binary data array of same shape as array which describes whether 
        individual samples in a variable (or all samples) are continuous 
        or discrete: 0s for continuous variables and 1s for discrete variables.
    self.missing_flag:
        Is missing_flag
    self.var_names:
        If var_names is not None:
            Is var_names
        If var_names is None:
            Is {i: i for i in range(self.N)}
    self.datatime : dictionary
        Time axis for each of the multiple datasets.
    self.analysis_mode : string
        Is analysis_mode
    self.reference_points: array-like
        If reference_points is not None:
            1D numpy array holding all specified reference_points, less those
            smaller than 0 and larger than self.largest_time_step-1
        If reference_points is None:
            Is np.array(self.largest_time_step)
    self.time_offsets : dictionary
        If time_offsets is not None:
            Is time_offsets
        If time_offsets is None:
            Is {key: 0 for key in self.values.keys()}
    self.M : int
        Number of datasets
    self.N : int
        Number of variables (constant across datasets)
    self.T : dictionary
        Dictionary {key(m): T(m), ...}, where T(m) is the time length of
        datasets m and key(m) its identifier as in self.values
    self.largest_time_step : int
        max_{0 <= m <= M} [ T(m) + time_offset(m)], i.e., the largest (latest)
        time step relative to the shared time axis for which at least one
        observation exists in the dataset.
    self.bootstrap : dictionary
        Whether to use bootstrap. Must be a dictionary with keys random_state,
        boot_samples, and boot_blocklength.
    """

    def __init__(self, data, mask=None, missing_flag=None, vector_vars=None, var_names=None,
        data_type=None, datatime=None, analysis_mode ='single', reference_points=None,
        time_offsets=None, remove_missing_upto_maxlag=False):

        # Check that a valid analysis mode, specified by the argument
        # 'analysis_mode', has been chosen
        if analysis_mode in ['single', 'multiple']:
            self.analysis_mode = analysis_mode
        else:
            raise ValueError("'analysis_mode' is '{}', must be 'single' or "\
                "'multiple'.".format(analysis_mode))

        # Check for correct type and format of 'data', internally cast to the
        # analysis mode 'multiple' case in dictionary representation
        if self.analysis_mode == 'single':
            # In this case the 'time_offset' functionality must not be used
            if time_offsets is not None:
                raise ValueError("'time_offsets' must be None in analysis "\
                    "mode'single'.")

            # 'data' must be either
            # - np.ndarray of shape (T, N)
            # - np.ndarray of shape (1, T, N)
            # - a dictionary with one element whose value is a np.ndarray of
            # shape (T, N)
            
            if isinstance(data, np.ndarray):
                _data_shape = data.shape
                if len(_data_shape) == 2:
                    self.values = {0: np.copy(data)}
                    self._initialized_from = "2d numpy array"
                elif len(_data_shape) == 3 and _data_shape[0] == 1:
                    self.values = {0: np.copy(data[0, :, :])}
                    self._initialized_from = "3d numpy array"
                else:
                    raise TypeError("In analysis mode 'single', 'data' given "\
                        "as np.ndarray. 'data' is of shape {}, must be of "\
                        "shape (T, N) or (1, T, N).".format(_data_shape))

            elif isinstance(data, dict):
                if len(data) == 1:
                    _data = next(iter(data.values()))
                    if isinstance(_data, np.ndarray):
                        if len(_data.shape) == 2:
                            self.values = data.copy()
                            self._initialized_from = "dict"
                        else:
                            raise TypeError("In analysis mode 'single', "\
                                "'data'given as dictionary. The single value "\
                                "is a np.ndarray of shape {}, must be of "\
                                "shape (T, N).".format(_data.shape))
                    else:
                        raise TypeError("In analysis mode 'single', 'data' "\
                            "given as dictionary. The single value is of type "\
                            "{}, must be np.ndarray.".format(type(_data)))

                else:
                    raise ValueError("In analysis mode 'single', 'data' given "\
                        "as dictionary. There are {} entries in 'data', there "\
                        "must be exactly one entry.".format(len(data)))

            else:
                raise TypeError("In analysis mode 'single'. 'data' is of type "\
                    "{}, must be np.ndarray or dict.".format(type(data)))

        elif self.analysis_mode == 'multiple':
            # 'data' must either be a
            # - np.ndarray of shape (M, T, N)
            # - dict whose values of are np.ndarray of shape (T_i, N), where T_i
            # may vary across the values

            if isinstance(data, np.ndarray):
                _data_shape = data.shape
                if len(_data_shape) == 3:
                    self.values = {i: np.copy(data[i, :, :]) for i in range(_data_shape[0])}
                    self._initialized_from = "3d numpy array"
                else:
                    raise TypeError("In analysis mode 'multiple', 'data' "\
                        "given as np.ndarray. 'data' is of shape {}, must be "\
                        "of shape (M, T, N).".format(_data_shape))

                # In this case the 'time_offset' functionality must not be used
                if time_offsets is not None:
                    raise ValueError("In analysis mode 'multiple'. Since "\
                        "'data' is given as np.ndarray, 'time_offsets' must "\
                        "be None.")

            elif isinstance(data, dict):
                _N_list = set()
                for dataset_key, dataset_data in data.items():
                    if isinstance(dataset_data, np.ndarray):
                        _dataset_data_shape = dataset_data.shape
                        if len(_dataset_data_shape) == 2:
                            _N_list.add(_dataset_data_shape[1])
                        else:
                            raise TypeError("In analysis mode 'multiple', "\
                                "'data' given as dictionary. 'data'[{}] is of "\
                                "shape {}, must be of shape (T_i, N).".format(
                                    dataset_key, _dataset_data_shape))

                    else:
                        raise TypeError("In analysis mode 'multiple', 'data' "\
                            "given as dictionary. 'data'[{}] is of type {}, "\
                            "must be np.ndarray.".format(dataset_key,
                                type(dataset_data)))

                if len(_N_list) == 1:
                    self.values = data.copy()
                    self._initialized_from = "dict"
                else:
                    raise ValueError("In analysis mode 'multiple', 'data' "\
                        "given as dictionary. All entries must be np.ndarrays "\
                        "of shape (T_i, N), where T_i may vary across the "\
                        "entries while N must not vary. In the given 'data' N "\
                        "varies.")

            else:
                raise TypeError("In analysis mode 'multiple'. 'data' is of "\
                    "type {}, must be np.ndarray or dict.".format(type(data)))

        # Store the keys of the datasets in a separated attribute
        self.datasets = list(self.values.keys())

        # Save the data format and check for NaNs:
        self.M = len(self.values) # (Number of datasets)

        self.T = dict() # (Time lengths of the individual datasets)
        for dataset_key, dataset_data in self.values.items():
            if np.isnan(dataset_data).sum() != 0:
                raise ValueError("NaNs in the data.")

            _dataset_data_shape = dataset_data.shape
            self.T[dataset_key] = _dataset_data_shape[0]
            self.Ndata = _dataset_data_shape[1] # (Number of variables) 
            # N does not vary across the datasets

        # Setup dictionary of variables for vector mode
        self.vector_vars = vector_vars
        if self.vector_vars is None:
            self.vector_vars = dict(zip(range(self.Ndata), [[(i, 0)] 
                                for i in range(self.Ndata)]))
            self.has_vector_data = False
        else:
            self.has_vector_data = True


        # TODO: check vector_vars!
        self.N = len(self.vector_vars)

        # Warnings
        if self.analysis_mode == 'single' and self.N > next(iter(self.T.values())):
            warnings.warn("In analysis mode 'single', 'data'.shape = ({}, {});"\
                " is it of shape (observations, variables)?".format(self.T[0],
                    self.N))

        if self.analysis_mode == 'multiple' and self.M == 1:
            warnings.warn("In analysis mode 'multiple'. There is just a "\
                "single dataset, is this as intended?'")


        # Save the variable names. If unspecified, use the default
        if var_names is None:
            self.var_names = {i: i for i in range(self.N)}
        else:
            self.var_names = var_names

        self.mask = None
        if mask is not None:
            self.mask = self._check_mask(mask = mask)
            
        self.data_type = None
        if data_type is not None:
            self.data_type = self._check_mask(mask = data_type)

        # Check and prepare the time offsets
        self._check_and_set_time_offsets(time_offsets)
        self.time_offsets_is_none = time_offsets is None

        # Set the default datatime if unspecified
        if datatime is None:
            self.datatime = {m: np.arange(self.time_offsets[m], 
                self.time_offsets[m] + self.T[m]) for m in self.values.keys()}
        else:
            if not isinstance(datatime, dict):
                self.datatime = {0: datatime}   
            else:
                self.datatime = datatime

        # Save the largest/smallest relevant time step
        self.largest_time_step = np.add(np.asarray(list(self.T.values())), np.asarray(list(self.time_offsets.values()))).max()
        self.smallest_time_step = np.add(np.asarray(list(self.T.values())), np.asarray(list(self.time_offsets.values()))).min()

        # Check and prepare the reference points
        self._check_and_set_reference_points(reference_points)
        self.reference_points_is_none = reference_points is None

        # Save the 'missing_flag' value
        self.missing_flag = missing_flag
        if self.missing_flag is not None:
            for dataset_key in self.values:
                self.values[dataset_key][self.values[dataset_key] == self.missing_flag] = np.nan
        self.remove_missing_upto_maxlag = remove_missing_upto_maxlag

        # If PCMCI.run_bootstrap_of is called, then the
        # bootstrap random draw can be set here
        self.bootstrap = None


    def _check_mask(self, mask):
        """Checks that the mask and data_type arrays:
            * have same shape as the data
            * are an numpy ndarray (or subtype)
            * do not contain any NaN entries
            * contain only 0 and 1

        """
        # Check that there is a mask if required
        _use_mask = mask

        # If we have a mask, check it
        if _use_mask is not None:
            # Check data type and generic format of 'mask', map to multiple datasets mode
            # dictionary representation
            if isinstance(_use_mask, np.ndarray):
                if len(_use_mask.shape) == 2:
                    _use_mask_dict = {0: _use_mask}
                elif len(_use_mask.shape) == 3:
                    if _use_mask.shape[0] == self.M:
                        _use_mask_dict = {i: _use_mask[i, :, :] for i in range(self.M)}
                    else:
                        raise ValueError("Shape mismatch: {} datasets "\
                            " in data but {} in (type) mask, must be "\
                            "identical.".format(self.M, _use_mask.shape[0]))

                else:
                    raise TypeError("data given as 3d np.ndarray. "\
                        "(type) mask is np.ndarray of shape {}, must be of "\
                        "shape (M, T, N).".format(_use_mask.shape))

            elif isinstance(_use_mask, dict):
                if len(_use_mask) == self.M:
                    for dataset_key in self.values.keys():
                        if _use_mask.get(dataset_key) is None:
                            raise ValueError("data has key {} (type {}) "\
                                "but (type) mask does not, keys must be "\
                                "identical.".format(dataset_key,
                                    type(dataset_key)))

                    _use_mask_dict = _use_mask

                else:
                    raise ValueError("Shape mismatch: {} datasets "\
                        "in data but {} in (type) mask, must be "\
                        "identical.".format(self.M, len(_use_mask)))
            else:
                raise TypeError("(type) mask is of type "\
                    "{}, must be dict or array.".format(type(_use_mask)))

            # Check for consistency with shape of 'self.values' and for NaNs
            for dataset_key, dataset_data in self.values.items():
                _use_mask_dict_data = _use_mask_dict[dataset_key] 
                if _use_mask_dict_data.shape == dataset_data.shape:
                    if np.sum(np.isnan(_use_mask_dict_data)) != 0:
                        raise ValueError("NaNs in the (type) data mask")
                    # if check_data_type:
                    if not set(np.unique(_use_mask_dict_data)).issubset(set([0, 1])):
                        raise ValueError("(Type) mask contains other values than 0 and 1")
                else:
                    if self.analysis_mode == 'single':
                        raise ValueError("Shape mismatch: data is of shape "\
                            "{}, (type) mask is of shape {}. Must be "\
                            "identical.".format(dataset_data.shape,
                                _use_mask_dict_data.shape))
                    elif self.analysis_mode == 'multiple':
                        raise ValueError("Shape mismatch: dataset {} "\
                            "is of shape {} in data and of shape {} in "\
                            "(type) mask. Must be identical.".format(dataset_key,
                                dataset_data.shape,
                                _use_mask_dict_data.shape))

            # Return the mask in dictionary format
            return _use_mask_dict

    def _check_and_set_time_offsets(self, time_offsets):
        """Check the argument 'time_offsets' for consistency and bring into
        canonical format"""

        if time_offsets is not None:

            assert self.analysis_mode == 'multiple'
            assert self._initialized_from == 'dict'

            # Check data type and generic format of 'time_offsets', map to
            # dictionary representation
            if isinstance(time_offsets, dict):
                if len(time_offsets) == self.M:
                    for dataset_key in self.values.keys():
                        if time_offsets.get(dataset_key) is None:
                            raise ValueError("'data' has key {} (type {}) but "\
                                "'time_offsets' does not, keys must be "\
                                "identical.".format(dataset_key,
                                    type(dataset_key)))

                    self.time_offsets = time_offsets

                else:
                    raise ValueError("Shape mismatch: {} datasets in "\
                        "'data' but {} in 'time_offsets', must be "\
                        "identical.".format(self.M, len(time_offsets)))

            else:
                raise TypeError("'time_offsets' is of type {}, must be "\
                    "dict.".format(type(time_offsets)))

            # All time offsets must be non-negative integers, at least one of
            # which is zero
            found_zero_time_offset = False
            for time_offset in self.time_offsets.values():
                if np.issubdtype(type(time_offset), np.integer):
                    if time_offset >= 0:
                        if time_offset == 0:
                            found_zero_time_offset = True
                    else:
                        raise ValueError("A dataset has time offset "\
                            "{}, must be non-negative.".format(time_offset))

                else:
                    raise TypeError("There is a time offset of type {}, must "\
                        "be int.".format(type(time_offset)))

            if not found_zero_time_offset:
                raise ValueError("At least one time offset must be 0.")

        else:
            # If no time offsets are specified, all of them are zero
            self.time_offsets = {dataset_key: 0 for dataset_key in self.values.keys()}

    def _check_and_set_reference_points(self, reference_points):
        """Check the argument 'reference_point' for consistency and bring into
        canonical format"""

        # Check type of 'reference_points' and its elements
        if reference_points is None:
            # If no reference point is specified, use as many reference points
            # as possible
            self.reference_points = np.arange(self.largest_time_step)

        elif isinstance(reference_points, int):
            # If a single reference point is specified as an int, convert it to
            # a single element numpy array
            self.reference_points = np.array([reference_points])

        elif isinstance(reference_points, np.ndarray):
            # Check that all reference points are ints
            for ref_point in reference_points:
                if not np.issubdtype(type(ref_point), np.integer):
                    raise TypeError("All reference points must be integers.")

            self.reference_points = reference_points

        elif isinstance(reference_points, list):
            # Check that all reference points are ints
            for ref_point in reference_points:
                if not isinstance(ref_point, int):
                    raise TypeError("All reference points must be integers.")

            # If given as a list, cast to numpy array
            self.reference_points = np.asarray(reference_points)

        else:
            raise TypeError("Unsupported data type of 'reference_points': Is "\
                "{}, must be None or int or a list or np.ndarray of "\
                "ints.".format(type(reference_points)))

        # Remove negative reference points
        if np.sum(self.reference_points < 0) > 0:
            warnings.warn("Some reference points were negative. These are "\
                "removed.")
            self.reference_points = self.reference_points[self.reference_points >= 0]

        # Remove reference points that are larger than the largest time step
        if np.sum(self.reference_points >= self.largest_time_step) > 0:
            warnings.warn("Some reference points were larger than the largest "\
                "relevant time step, which here is {}. These are "\
                "removed.".format(self.largest_time_step - 1))
            self.reference_points = self.reference_points[self.reference_points < self.largest_time_step]

        # Raise an error if no valid reference points was specified
        if len(self.reference_points) == 0:
            raise ValueError("No valid reference point.") 


    def construct_array(self, X, Y, Z, tau_max,
                        extraZ=None,
                        mask=None,
                        mask_type=None,
                        data_type=None,
                        return_cleaned_xyz=False,
                        do_checks=True,
                        remove_overlaps=True,
                        cut_off='2xtau_max',
                        verbosity=0):
        """Constructs array from variables X, Y, Z from data.
        Data is of shape (T, N) if analysis_mode == 'single', where T is the
        time series length and N the number of variables, and of (n_ens, T, N)
        if analysis_mode == 'multiple'.

        Parameters
        ----------
        X, Y, Z, extraZ : list of tuples
            For a dependence measure I(X;Y|Z), X, Y, Z can be multivariate of
            the form [(var1, -lag), (var2, -lag), ...]. At least one varlag in Y 
            has to be at lag zero. extraZ is only used in CausalEffects class.
        tau_max : int
            Maximum time lag. This may be used to make sure that estimates for
            different lags in X and Z all have the same sample size.
        mask : array-like, optional (default: None)
            Optional mask array, must be of same shape as data.  If it is set,
            then it overrides the self.mask assigned to the dataframe. If it is
            None, then the self.mask is used, if it exists.
        mask_type : {None, 'y','x','z','xy','xz','yz','xyz'}
            Masking mode: Indicators for which variables in the dependence
            measure I(X; Y | Z) the samples should be masked. If None, the mask
            is not used. Explained in tutorial on masking and missing values.
        data_type : array-like
            Binary data array of same shape as array which describes whether 
            individual samples in a variable (or all samples) are continuous 
            or discrete: 0s for continuous variables and 1s for discrete variables.
            If it is set, then it overrides the self.data_type assigned to the dataframe.
        return_cleaned_xyz : bool, optional (default: False)
            Whether to return cleaned X,Y,Z, where possible duplicates are
            removed.
        do_checks : bool, optional (default: True)
            Whether to perform sanity checks on input X,Y,Z
        remove_overlaps : bool, optional (default: True)
            Whether to remove variables from Z/extraZ if they overlap with X or Y.
        cut_off : {'2xtau_max', 'tau_max', 'max_lag', 'max_lag_or_tau_max', 2xtau_max_future}
            If cut_off == '2xtau_max':
                - 2*tau_max samples are cut off at the beginning of the time
                  series ('beginning' here refers to the temporally first
                  time steps). This guarantees that (as long as no mask is
                  used) all MCI tests are conducted on the same samples,
                  independent of X, Y, and Z.

                - If at time step t_missing a data value is missing, then the
                  time steps t_missing, ..., t_missing + 2*tau_max are cut
                  out. The latter part only holds if
                  remove_missing_upto_maxlag=True.

            If cut_off ==  'max_lag':
                - max_lag(X, Y, Z) samples are cut off at the beginning of the
                  time series, where max_lag(X, Y, Z) is the maximum lag of
                  all nodes in X, Y, and Z. These are all samples that can in
                  principle be used.

                - If at time step t_missing a data value is missing, then the
                  time steps t_missing, ..., t_missing + max_lag(X, Y, Z) are
                  cut out. The latter part only holds if
                  remove_missing_upto_maxlag=True.

            If cut_off == 'max_lag_or_tau_max':
                - max(max_lag(X, Y, Z), tau_max) are cut off at the beginning.
                  This may be useful for modeling by comparing multiple
                  models on the same samples. 

                - If at time step t_missing a data value is missing, then the
                  time steps t_missing, ..., t_missing + max(max_lag(X, Y,
                  Z), tau_max) are cut out. The latter part only holds if
                  remove_missing_upto_maxlag=True.

            If cut_off == 'tau_max':
                - tau_max samples are cut off at the beginning. This may be
                  useful for modeling by comparing multiple models on the
                  same samples. 

                - If at time step t_missing a data value is missing, then the
                  time steps t_missing, ..., t_missing + max(max_lag(X, Y,
                  Z), tau_max) are cut out. The latter part only holds if
                  remove_missing_upto_maxlag=True.
                
            If cut_off == '2xtau_max_future':
                First, the relevant time steps are determined as for cut_off ==
                'max_lag'. Then, the temporally latest time steps are removed
                such that the same number of time steps remains as there would
                be for cut_off == '2xtau_max'. This may be useful when one is
                mostly interested in the temporally first time steps and would
                like all MCI tests to be performed on the same *number* of
                samples. Note, however, that while the *number* of samples is
                the same for all MCI tests, the samples themselves may be
                different.
        verbosity : int, optional (default: 0)
            Level of verbosity.

        Returns
        -------
        array, xyz [,XYZ], data_type : Tuple of data array of shape (dim, n_samples),
            xyz identifier array of shape (dim,) identifying which row in array
            corresponds to X, Y, and Z, and the type mask that indicates which samples
            are continuous or discrete. For example: X = [(0, -1)],
            Y = [(1, 0)], Z = [(1, -1), (0, -2)] yields an array of shape
            (4, n_samples) and xyz is xyz = numpy.array([0,1,2,2]). If
            return_cleaned_xyz is True, also outputs the cleaned XYZ lists.
        """

        # # This version does not yet work with bootstrap
        # try:
        #     assert self.bootstrap is None
        # except AssertionError:
        #     print("This version does not yet work with bootstrap.")
        #     raise

        if extraZ is None:
            extraZ = []
        
        if Z is None:
            Z = []

        # If vector-valued variables exist, add them
        def vectorize(varlag):     
            vectorized_var = []
            for (var, lag) in varlag:
                for (vector_var, vector_lag) in self.vector_vars[var]:
                    vectorized_var.append((vector_var, vector_lag + lag))
            return vectorized_var



        # if extended_summary_graph_lag is not None:
        #     assert self.vector_vars does not contain lagged microvariables # only space-vectors
        #     assert tau_max == 1  # or 0 if mode is 'summary_graph' /?
        #     # Y = vectorize(Y)
        #     varX, lagX = X[0]  # because X = [(i, -lag)]
            
        #     if lagX == 0:
        #         pass #X = vectorize(X)
        #     elif: lagX == -1:
        #         X = [(varX, -lag) for lag in range(1, extended_summary_graph_lag + 1)]
        #     else:
        #         raise ValueError("Extended summary graph can only have tau_max = 1")

        #     Znew = []
        #     for z in Z:
        #         varZ, lagZ = z   # z = (k, -1) or (k, -2) or (k, 0)
        #         if: lagZ == 0:
        #             Znew += z
        #         elif: lagZ == -1:
        #             Znew += [(varZ, -lag) for lag in range(1, extended_summary_graph_lag + 1)]   
        #         elif: lagZ == -2:
        #             Znew += [(varZ, -lag) for lag in range(2, 2*extended_summary_graph_lag + 1)]   
        #         else:
        #             raise ValueError("Extended summary graph can only have tau_max = 1")
        #     Z = Znew


        X = vectorize(X) 
        Y = vectorize(Y) 
        Z = vectorize(Z) 
        extraZ = vectorize(extraZ) 

        # Remove duplicates in X, Y, Z, extraZ
        X = list(OrderedDict.fromkeys(X))
        Y = list(OrderedDict.fromkeys(Y))
        Z = list(OrderedDict.fromkeys(Z))
        extraZ = list(OrderedDict.fromkeys(extraZ))

        if remove_overlaps:
            # If a node in Z occurs already in X or Y, remove it from Z
            Z = [node for node in Z if (node not in X) and (node not in Y)]
            extraZ = [node for node in extraZ if (node not in X) and (node not in Y) and (node not in Z)]

        XYZ = X + Y + Z + extraZ
        dim = len(XYZ)

        # Check that all lags are non-positive and indices are in [0,N-1]
        if do_checks:
            self._check_nodes(Y, XYZ, self.Ndata, dim)

        # Use the mask, override if needed
        _mask = mask
        if _mask is None:
            _mask = self.mask
        else:
            _mask = self._check_mask(mask = _mask)
            
        _data_type = data_type
        if _data_type is None:
            _data_type = self.data_type
        else:
            _data_type = self._check_mask(mask = _data_type, check_data_type=True)

        # Figure out what cut off we will be using
        if cut_off == '2xtau_max':
            max_lag = 2*tau_max
        elif cut_off == 'max_lag':
            max_lag = abs(np.array(XYZ)[:, 1].min())
        elif cut_off == 'tau_max':
            max_lag = tau_max
        elif cut_off == 'max_lag_or_tau_max':
            max_lag = max(abs(np.array(XYZ)[:, 1].min()), tau_max)
        elif cut_off == '2xtau_max_future':
            ## TODO: CHECK THIS
            max_lag = abs(np.array(XYZ)[:, 1].min())
        else:
            raise ValueError("max_lag must be in {'2xtau_max', 'tau_max', 'max_lag', "\
                "'max_lag_or_tau_max', '2xtau_max_future'}")

        # Setup XYZ identifier
        index_code = {'x' : 0,
                      'y' : 1,
                      'z' : 2,
                      'e' : 3}
        xyz = np.array([index_code[name]
                        for var, name in zip([X, Y, Z, extraZ], ['x', 'y', 'z', 'e'])
                        for _ in var])

        # Run through all datasets and fill a dictionary holding the
        # samples taken from the individual datasets
        samples_datasets = dict()
        data_types = dict()
        self.use_indices_dataset_dict = dict()

        for dataset_key, dataset_data in self.values.items():

            # Apply time offset to the reference points
            ref_points_here = self.reference_points - self.time_offsets[dataset_key]

            # Remove reference points that are out of bounds or are to be
            # excluded given the choice of 'cut_off'
            ref_points_here = ref_points_here[ref_points_here >= max_lag]
            ref_points_here = ref_points_here[ref_points_here < self.T[dataset_key]]

            # Keep track of which reference points would have remained for
            # max_lag == 2*tau_max
            if cut_off == '2xtau_max_future':
                ref_points_here_2_tau_max = self.reference_points - self.time_offsets[dataset_key]
                ref_points_here_2_tau_max = ref_points_here_2_tau_max[ref_points_here_2_tau_max  >= 2*tau_max]
                ref_points_here_2_tau_max = ref_points_here_2_tau_max[ref_points_here_2_tau_max  < self.T[dataset_key]]

            # Sort the valid reference points (not needed, but might be useful
            # for detailed debugging)
            ref_points_here = np.sort(ref_points_here)

            # For cut_off == '2xtau_max_future' reduce the samples size the
            # number of samples that would have been obtained for cut_off ==
            # '2xtau_max', removing the temporally latest ones
            if cut_off == '2xtau_max_future':
                n_to_cut_off = len(ref_points_here) - len(ref_points_here_2_tau_max)
                assert n_to_cut_off >= 0
                if n_to_cut_off > 0:
                    ref_points_here = np.sort(ref_points_here)
                    ref_points_here = ref_points_here[:-n_to_cut_off]

            # If no valid reference points are left, continue with the next dataset
            if len(ref_points_here) == 0:
                continue

            if self.bootstrap is not None:

                boot_blocklength = self.bootstrap['boot_blocklength']

                if boot_blocklength == 'cube_root':
                    boot_blocklength = max(1, int(len(ref_points_here)**(1/3)))
                # elif boot_blocklength == 'from_autocorrelation':
                #     boot_blocklength = \
                #         get_block_length(overlapping_residuals.T, xyz=np.zeros(N), mode='confidence')
                elif type(boot_blocklength) is int and boot_blocklength > 0:
                    pass
                else:
                    raise ValueError("boot_blocklength must be integer > 0, 'cube_root', or 'from_autocorrelation'")

                # Chooses THE SAME random seed for every dataset, maybe that's what we want...
                # If the reference points are all the same, this will give the same bootstrap
                # draw. However, if they are NOT the same, they will differ. 
                # TODO: Decide whether bootstrap draws should be the same for each dataset and
                # how to achieve that if the reference points differ...
                # random_state = self.bootstrap['random_state']
                random_state = deepcopy(self.bootstrap['random_state'])

                # Determine the number of blocks total, rounding up for non-integer
                # amounts
                n_blks = int(math.ceil(float(len(ref_points_here))/boot_blocklength))

                if n_blks < 10:
                    raise ValueError("Only %d block(s) for block-sampling,"  %n_blks +
                                     " choose smaller boot_blocklength!")

                # Get the starting indices for the blocks
                blk_strt = random_state.choice(np.arange(len(ref_points_here) - boot_blocklength), size=n_blks, replace=True)
                # Get the empty array of block resampled values
                boot_draw = np.zeros(n_blks*boot_blocklength, dtype='int')
                # Fill the array of block resamples
                for i in range(boot_blocklength):
                    boot_draw[i::boot_blocklength] = ref_points_here[blk_strt + i]
                # Cut to proper length
                ref_points_here = boot_draw[:len(ref_points_here)]

            # Construct the data array holding the samples taken from the
            # current dataset
            samples_datasets[dataset_key] = np.zeros((dim, len(ref_points_here)), dtype = dataset_data.dtype)
            for i, (var, lag) in enumerate(XYZ):
                samples_datasets[dataset_key][i, :] = dataset_data[ref_points_here + lag, var]

            # Build the mask array corresponding to this dataset
            if _mask is not None:
                mask_dataset = np.zeros((dim, len(ref_points_here)), dtype = 'bool')
                for i, (var, lag) in enumerate(XYZ):
                    mask_dataset[i, :] = _mask[dataset_key][ref_points_here + lag, var]

            # Take care of masking
            use_indices_dataset = np.ones(len(ref_points_here), dtype = 'int')

            # Build the type mask array corresponding to this dataset
            if _data_type is not None:
                data_type_dataset = np.zeros((dim, len(ref_points_here)), dtype = 'bool')
                for i, (var, lag) in enumerate(XYZ):
                    data_type_dataset[i, :] = _data_type[dataset_key][ref_points_here + lag, var]
                data_types[dataset_key] = data_type_dataset
            
            # Remove all values that have missing value flag, and optionally as well the time
            # slices that occur up to max_lag after
            if self.missing_flag is not None:
                missing_anywhere = np.array(np.where(np.any(np.isnan(samples_datasets[dataset_key]), axis=0))[0])

                if self.remove_missing_upto_maxlag:
                    idx_to_remove = set(idx + tau for idx in missing_anywhere for tau in range(max_lag + 1))
                else:
                    idx_to_remove = set(idx for idx in missing_anywhere)
                
                use_indices_dataset[np.array(list(idx_to_remove), dtype='int')] = 0
            
            if _mask is not None:
                # Remove samples with mask == 1 conditional on which mask_type
                # is used

                # Iterate over defined mapping from letter index to number index,
                # i.e. 'x' -> 0, 'y' -> 1, 'z'-> 2, 'e'-> 3
                for idx, cde in index_code.items():
                    # Check if the letter index is in the mask type
                    if (mask_type is not None) and (idx in mask_type):
                        # If so, check if any of the data that correspond to the
                        # letter index is masked by taking the product along the
                        # node-data to return a time slice selection, where 0
                        # means the time slice will not be used
                        slice_select = np.prod(mask_dataset[xyz == cde, :] == False, axis=0)
                        use_indices_dataset *= slice_select

            # Accordingly update the data array and data type array
            samples_datasets[dataset_key] = samples_datasets[dataset_key][:, use_indices_dataset == 1]
            if _data_type is not None:
                data_types[dataset_key] = data_types[dataset_key][:, use_indices_dataset == 1]

        ## end for dataset_key, dataset_data in self.values.items()

        # Save used indices as attribute
        if len(ref_points_here) > 0:
            self.use_indices_dataset_dict[dataset_key] = ref_points_here[use_indices_dataset==1]
        else:
            self.use_indices_dataset_dict[dataset_key] = []

        # Concatenate the arrays of all datasets
        array = np.concatenate(tuple(samples_datasets.values()), axis = 1)
        if _data_type is not None:
            type_array = np.concatenate(tuple(data_types.values()), axis = 1)
        else:
            type_array = None
        
        # print(np.where(np.isnan(array)))
        # print(array.shape)

        # Check whether there is any valid sample
        if array.shape[1] == 0:
            raise ValueError("No valid samples")

        # Print information about the constructed array
        if verbosity > 2:
            self.print_array_info(array, X, Y, Z, self.missing_flag, mask_type, type_array, extraZ)

        # Return the array and xyz and optionally (X, Y, Z)
        if return_cleaned_xyz:
            return array, xyz, (X, Y, Z), type_array

        return array, xyz, type_array

    def _check_nodes(self, Y, XYZ, N, dim):
        """
        Checks that:
        * The requests XYZ nodes have the correct shape
        * All lags are non-positive
        * All indices are less than N
        * One of the Y nodes has zero lag

        Parameters
        ----------
        Y : list of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.
        XYZ : list of tuples
            List of nodes chosen for current independence test
        N : int
            Total number of listed nodes
        dim : int
            Number of nodes excluding repeated nodes
        """
        if np.array(XYZ).shape != (dim, 2):
            raise ValueError("X, Y, Z must be lists of tuples in format"
                             " [(var, -lag),...], eg., [(2, -2), (1, 0), ...]")
        if np.any(np.array(XYZ)[:, 1] > 0):
            raise ValueError("nodes are %s, " % str(XYZ) +
                             "but all lags must be non-positive")
        if (np.any(np.array(XYZ)[:, 0] >= N)
                or np.any(np.array(XYZ)[:, 0] < 0)):
            raise ValueError("var indices %s," % str(np.array(XYZ)[:, 0]) +
                             " but must be in [0, %d]" % (N - 1))
        # if np.all(np.array(Y)[:, 1] != 0):
        #     raise ValueError("Y-nodes are %s, " % str(Y) +
        #                      "but one of the Y-nodes must have zero lag")

    def print_array_info(self, array, X, Y, Z, missing_flag, mask_type, data_type=None, extraZ=None):
        """
        Print info about the constructed array

        Parameters
        ----------
        array : Data array of shape (dim, T)
            Data array.
        X, Y, Z, extraZ : list of tuples
            For a dependence measure I(X;Y|Z), Y is of the form [(varY, 0)],
            where var specifies the variable index. X typically is of the form
            [(varX, -tau)] with tau denoting the time lag and Z can be
            multivariate [(var1, -lag), (var2, -lag), ...] .
        missing_flag : number, optional (default: None)
            Flag for missing values. Dismisses all time slices of samples where
            missing values occur in any variable and also flags samples for all
            lags up to 2*tau_max. This avoids biases, see section on masking in
            Supplement of [1]_.
        mask_type : {'y','x','z','xy','xz','yz','xyz'}
            Masking mode: Indicators for which variables in the dependence
            measure I(X; Y | Z) the samples should be masked. If None, the mask
            is not used. Explained in tutorial on masking and missing values.
        data_type : array-like
            Binary data array of same shape as array which describes whether 
            individual samples in a variable (or all samples) are continuous 
            or discrete: 0s for continuous variables and 1s for discrete variables.
        """
        if extraZ is None:
            extraZ = []
        indt = " " * 12
        print(indt + "Constructed array of shape %s from"%str(array.shape) +
              "\n" + indt + "X = %s" % str(X) +
              "\n" + indt + "Y = %s" % str(Y) +
              "\n" + indt + "Z = %s" % str(Z))
        if extraZ is not None:  
            print(indt + "extraZ = %s" % str(extraZ))
        if self.mask is not None and mask_type is not None:
            print(indt+"with masked samples in %s removed" % mask_type)
        if self.data_type is not None:
            print(indt+"with %s % discrete values" % np.sum(data_type)/data_type.size)
        if self.missing_flag is not None:
            print(indt+"with missing values = %s removed" % self.missing_flag)


def get_acf(series, max_lag=None):
    """Returns autocorrelation function.

    Parameters
    ----------
    series : 1D-array
        data series to compute autocorrelation from

    max_lag : int, optional (default: None)
        maximum lag for autocorrelation function. If None is passed, 10% of
        the data series length are used.

    Returns
    -------
    autocorr : array of shape (max_lag + 1,)
        Autocorrelation function.
    """
    # Set the default max lag
    if max_lag is None:
        max_lag = int(max(5, 0.1*len(series)))
    # Initialize the result
    autocorr = np.ones(max_lag + 1)
    # Iterate over possible lags
    for lag in range(1, max_lag + 1):
        # Set the values
        y1_vals = series[lag:]
        y2_vals = series[:len(series) - lag]
        # Calculate the autocorrelation
        autocorr[lag] = np.corrcoef(y1_vals, y2_vals, ddof=0)[0, 1]
    return autocorr

def get_block_length(array, xyz, mode):
    """Returns optimal block length for significance and confidence tests.

    Determine block length using approach in Mader (2013) [Eq. (6)] which
    improves the method of Pfeifer (2005) with non-overlapping blocks In
    case of multidimensional X, the max is used. Further details in [1]_.
    Two modes are available. For mode='significance', only the indices
    corresponding to X are shuffled in array. For mode='confidence' all
    variables are jointly shuffled. If the autocorrelation curve fit fails,
    a block length of 5% of T is used. The block length is limited to a
    maximum of 10% of T.

    Mader et al., Journal of Neuroscience Methods,
    Volume 219, Issue 2, 15 October 2013, Pages 285-291

    Parameters
    ----------
    array : array-like
        data array with X, Y, Z in rows and observations in columns

    xyz : array of ints
        XYZ identifier array of shape (dim,).

    mode : str
        Which mode to use.

    Returns
    -------
    block_len : int
        Optimal block length.
    """
    # Inject a dependency on siganal, optimize
    from scipy import signal, optimize
    # Get the shape of the array
    dim, T = array.shape
    # Initiailize the indices
    indices = range(dim)
    if mode == 'significance':
        indices = np.where(xyz == 0)[0]

    # Maximum lag for autocov estimation
    max_lag = int(0.1*T)
    # Define the function to optimize against
    def func(x_vals, a_const, decay):
        return a_const * decay**x_vals

    # Calculate the block length
    block_len = 1
    for i in indices:
        # Get decay rate of envelope of autocorrelation functions
        # via hilbert trafo
        autocov = get_acf(series=array[i], max_lag=max_lag)
        autocov[0] = 1.
        hilbert = np.abs(signal.hilbert(autocov))
        # Try to fit the curve
        try:
            popt, _ = optimize.curve_fit(
                f=func,
                xdata=np.arange(0, max_lag+1),
                ydata=hilbert,
            )
            phi = popt[1]
            # Formula of Pfeifer (2005) assuming non-overlapping blocks
            l_opt = (4. * T * (phi / (1. - phi) + phi**2 / (1. - phi)**2)**2
                     / (1. + 2. * phi / (1. - phi))**2)**(1. / 3.)
            block_len = max(block_len, int(l_opt))
        except RuntimeError:
            warnings.warn("Error - curve_fit failed for estimating block_shuffle length, using"
                  " block_len = %d" % (int(.05 * T)))
            # block_len = max(int(.05 * T), block_len)
    # Limit block length to a maximum of 10% of T
    block_len = min(block_len, int(0.1 * T))
    return block_len


def lowhighpass_filter(data, cutperiod, pass_periods='low'):
    """Butterworth low- or high pass filter.

    This function applies a linear filter twice, once forward and once
    backwards. The combined filter has linear phase.

    Parameters
    ----------
    data : array
        Data array of shape (time, variables).
    cutperiod : int
        Period of cutoff.
    pass_periods : str, optional (default: 'low')
        Either 'low' or 'high' to act as a low- or high-pass filter

    Returns
    -------
    data : array
        Filtered data array.
    """
    try:
        from scipy.signal import butter, filtfilt
    except:
        print('Could not import scipy.signal for butterworth filtering!')

    fs = 1.
    order = 3
    ws = 1. / cutperiod / (0.5 * fs)
    b, a = butter(order, ws, pass_periods)
    if np.ndim(data) == 1:
        data = filtfilt(b, a, data)
    else:
        for i in range(data.shape[1]):
            data[:, i] = filtfilt(b, a, data[:, i])

    return data


def smooth(data, smooth_width, kernel='gaussian',
           mask=None, residuals=False, verbosity=0):
    """Returns either smoothed time series or its residuals.

    the difference between the original and the smoothed time series
    (=residuals) of a kernel smoothing with gaussian (smoothing kernel width =
    twice the sigma!) or heaviside window, equivalent to a running mean.

    Assumes data of shape (T, N) or (T,)
    :rtype: array
    :returns: smoothed/residual data

    Parameters
    ----------
    data : array
        Data array of shape (time, variables).
    smooth_width : float
        Window width of smoothing, 2*sigma for a gaussian.
    kernel : str, optional (default: 'gaussian')
        Smoothing kernel, 'gaussian' or 'heaviside' for a running mean.
    mask : bool array, optional (default: None)
        Data mask where True labels masked samples.
    residuals : bool, optional (default: False)
        True if residuals should be returned instead of smoothed data.
    verbosity : int, optional (default: 0)
        Level of verbosity.

    Returns
    -------
    data : array-like
        Smoothed/residual data.
    """

    if verbosity > 0:
        print("%s %s smoothing with " % ({True: "Take residuals of a ",
                                      False: ""}[residuals], kernel) +
          "window width %.2f (=2*sigma for a gaussian!)" % (smooth_width))

    totaltime = len(data)
    if kernel == 'gaussian':
        window = np.exp(-(np.arange(totaltime).reshape((1, totaltime)) -
                             np.arange(totaltime).reshape((totaltime, 1))
                             ) ** 2 / ((2. * smooth_width / 2.) ** 2))
    elif kernel == 'heaviside':
        import scipy.linalg
        wtmp = np.zeros(totaltime)
        wtmp[:int(np.ceil(smooth_width / 2.))] = 1
        window = scipy.linalg.toeplitz(wtmp)

    if mask is None:
        if np.ndim(data) == 1:
            smoothed_data = (data * window).sum(axis=1) / window.sum(axis=1)
        else:
            smoothed_data = np.zeros(data.shape)
            for i in range(data.shape[1]):
                smoothed_data[:, i] = (
                    data[:, i] * window).sum(axis=1) / window.sum(axis=1)
    else:
        if np.ndim(data) == 1:
            smoothed_data = ((data * window * (mask==False)).sum(axis=1) /
                             (window * (mask==False)).sum(axis=1))
        else:
            smoothed_data = np.zeros(data.shape)
            for i in range(data.shape[1]):
                smoothed_data[:, i] = ((
                    data[:, i] * window * (mask==False)[:, i]).sum(axis=1) /
                    (window * (mask==False)[:, i]).sum(axis=1))

    if residuals:
        return data - smoothed_data
    else:
        return smoothed_data


def weighted_avg_and_std(values, axis, weights):
    """Returns the weighted average and standard deviation.

    Parameters
    ---------
    values : array
        Data array of shape (time, variables).
    axis : int
        Axis to average/std about
    weights : array
        Weight array of shape (time, variables).

    Returns
    -------
    (average, std) : tuple of arrays
        Tuple of weighted average and standard deviation along axis.
    """

    values[np.isnan(values)] = 0.
    average = np.ma.average(values, axis=axis, weights=weights)

    variance = np.sum(weights * (values - np.expand_dims(average, axis)
                                    ) ** 2, axis=axis) / weights.sum(axis=axis)

    return (average, np.sqrt(variance))


def time_bin_with_mask(data, time_bin_length, mask=None):
    """Returns time binned data where only about non-masked values is averaged.

    Parameters
    ----------
    data : array
        Data array of shape (time, variables).
    time_bin_length : int
        Length of time bin.
    mask : bool array, optional (default: None)
        Data mask where True labels masked samples.

    Returns
    -------
    (bindata, T) : tuple of array and int
        Tuple of time-binned data array and new length of array.
    """

    T = len(data)

    time_bin_length = int(time_bin_length)

    if mask is None:
        sample_selector = np.ones(data.shape)
    else:
        # Invert mask
        sample_selector = (mask == False)

    if np.ndim(data) == 1.:
        data.shape = (T, 1)
        if mask is not None:
            mask.shape = (T, 1)
        else:
            sample_selector = np.ones(data.shape)

    bindata = np.zeros(
        (T // time_bin_length,) + data.shape[1:], dtype="float32")
    for index, i in enumerate(range(0, T - time_bin_length + 1,
                                    time_bin_length)):
        # print weighted_avg_and_std(fulldata[i:i+time_bin_length], axis=0,
        # weights=sample_selector[i:i+time_bin_length])[0]
        bindata[index] = weighted_avg_and_std(data[i:i + time_bin_length],
                                              axis=0,
                                              weights=sample_selector[i:i +
                                              time_bin_length])[0]

    T, grid_size = bindata.shape

    return (bindata.squeeze(), T)

def trafo2normal(data, mask=None, thres=0.001):
    """Transforms input data to standard normal marginals.

    Assumes data.shape = (T, dim)

    Parameters
    ----------
    data : array
        Data array of shape (time, variables).
    thres : float
        Set outer points in CDF to this value.
    mask : bool array, optional (default: None)
        Data mask where True labels masked samples.

    Returns
    -------
    normal_data : array-like
        data with standard normal marginals.
    """

    def trafo(xi):
        xisorted = np.sort(xi)
        yi = np.linspace(1. / len(xi), 1, len(xi))
        return np.interp(xi, xisorted, yi)

    normal_data = np.copy(data)

    if np.ndim(data) == 1:
        if mask is None:
            nonmasked = np.where(np.isnan(data) == False)[0]
        else:
            nonmasked = np.where((mask==0)*(np.isnan(data) == False))

        u = trafo(data[nonmasked])
        u[u==0.] = thres
        u[u==1.] = 1. - thres
        normal_data[nonmasked] = stats.norm.ppf(u)
    else:
        for i in range(data.shape[1]):
            if mask is None:
                nonmasked = np.where(np.isnan(data[:,i]) == False)[0]
            else:
                nonmasked = np.where((mask[:, i]==0)*(np.isnan(data[:, i]) == False))
                # nonmasked = np.where(mask[:, i]==0)
            # print(data[:, i].shape, nonmasked.shape)
            uniform = trafo(data[:, i][nonmasked])
            
            # print(data[-3:, i][nonmasked])

            uniform[uniform==0.] = thres
            uniform[uniform==1.] = 1. - thres
            normal_data[:, i][nonmasked] = stats.norm.ppf(uniform)

    return normal_data

# @jit(nopython=True)
def _get_patterns(array, array_mask, patt, patt_mask, weights, dim, step, fac, N, T):
    v = np.zeros(dim, dtype='float')

    start = step * (dim - 1)
    for n in range(0, N):
        for t in range(start, T):
            mask = 1
            ave = 0.
            for k in range(0, dim):
                tau = k * step
                v[k] = array[t - tau, n]
                ave += v[k]
                mask *= array_mask[t - tau, n]
            ave /= dim
            var = 0.
            for k in range(0, dim):
                var += (v[k] - ave) ** 2
            var /= dim
            weights[t - start, n] = var
            if (v[0] < v[1]):
                p = 1
            else:
                p = 0
            for i in range(2, dim):
                for j in range(0, i):
                    if (v[j] < v[i]):
                        p += fac[i]
            patt[t - start, n] = p
            patt_mask[t - start, n] = mask

    return patt, patt_mask, weights

def ordinal_patt_array(array, array_mask=None, dim=2, step=1,
                        weights=False, seed=None, verbosity=0):
    """Returns symbolified array of ordinal patterns.

    Each data vector (X_t, ..., X_t+(dim-1)*step) is converted to its rank
    vector. E.g., (0.2, -.6, 1.2) --> (1,0,2) which is then assigned to a
    unique integer (see Article). There are faculty(dim) possible rank vectors.

    Note that the symb_array is step*(dim-1) shorter than the original array!

    Reference: B. Pompe and J. Runge (2011). Momentary information transfer as
    a coupling measure of time series. Phys. Rev. E, 83(5), 1-12.
    doi:10.1103/PhysRevE.83.051122

    Parameters
    ----------
    array : array-like
        Data array of shape (time, variables).
    array_mask : bool array
        Data mask where True labels masked samples.
    dim : int, optional (default: 2)
        Pattern dimension
    step : int, optional (default: 1)
        Delay of pattern embedding vector.
    weights : bool, optional (default: False)
        Whether to return array of variances of embedding vectors as weights.
    seed : int
        For adding noise to break ties.
    verbosity : int, optional (default: 0)
        Level of verbosity.

    Returns
    -------
    patt, patt_mask [, patt_time] : tuple of arrays
        Tuple of converted pattern array and new length
    """
    random_state = np.random.default_rng(seed)

    from scipy.special import factorial

    array = array.astype('float64')

    if array_mask is not None:
        assert array_mask.dtype == 'int32'
    else:
        array_mask = np.zeros(array.shape, dtype='int32')


    if np.ndim(array) == 1:
        T = len(array)
        array = array.reshape(T, 1)
        array_mask = array_mask.reshape(T, 1)

    # Add noise to destroy ties...
    array += (1E-6 * array.std(axis=0)
              * random_state.random((array.shape[0], array.shape[1])).astype('float64'))

    patt_time = int(array.shape[0] - step * (dim - 1))
    T, N = array.shape

    if dim <= 1 or patt_time <= 0:
        raise ValueError("Dim mist be > 1 and length of delay vector smaller "
                         "array length.")

    patt = np.zeros((patt_time, N), dtype='int32')
    weights_array = np.zeros((patt_time, N), dtype='float64')

    patt_mask = np.zeros((patt_time, N), dtype='int32')

    # Precompute factorial for c-code... patterns of dimension
    # larger than 10 are not supported
    fac = factorial(np.arange(10)).astype('int32')

    # _get_patterns assumes mask=0 to be a masked value
    array_mask = (array_mask == False).astype('int32')

    (patt, patt_mask, weights_array) = _get_patterns(array, array_mask, patt, patt_mask, weights_array, dim, step, fac, N, T)

    weights_array = np.asarray(weights_array)
    patt = np.asarray(patt)
    # Transform back to mask=1 implying a masked value
    patt_mask = np.asarray(patt_mask) == False

    if weights:
        return patt, patt_mask, patt_time, weights_array
    else:
        return patt, patt_mask, patt_time


def quantile_bin_array(data, bins=6):
    """Returns symbolified array with equal-quantile binning.

    Parameters
    ----------
    data : array
        Data array of shape (time, variables).
    bins : int, optional (default: 6)
        Number of bins.

    Returns
    -------
    symb_array : array
        Converted data of integer type.
    """
    T, N = data.shape

    # get the bin quantile steps
    bin_edge = int(np.ceil(T / float(bins)))

    symb_array = np.zeros((T, N), dtype='int32')

    # get the lower edges of the bins for every time series
    edges = np.sort(data, axis=0)[::bin_edge, :].T
    bins = edges.shape[1]

    # This gives the symbolic time series
    symb_array = (data.reshape(T, N, 1) >= edges.reshape(1, N, bins)).sum(
        axis=2) - 1

    return symb_array.astype('int32')


def var_process(parents_neighbors_coeffs, T=1000, use='inv_inno_cov',
                verbosity=0, initial_values=None):
    """Returns a vector-autoregressive process with correlated innovations.

    Wrapper around var_network with possibly more user-friendly input options.

    DEPRECATED. Will be removed in future.
    """
    print("data generating models are now in toymodels folder: "
         "from tigramite.toymodels import structural_causal_processes as toys.")
    return None

def structural_causal_process(links, T, noises=None, 
                        intervention=None, intervention_type='hard',
                        seed=None):
    """Returns a structural causal process with contemporaneous and lagged
    dependencies.

    DEPRECATED. Will be removed in future.
    """
    print("data generating models are now in toymodels folder: "
         "from tigramite.toymodels import structural_causal_processes as toys.")
    return None


if __name__ == '__main__':
    
    from tigramite.toymodels.structural_causal_processes import structural_causal_process
    ## Generate some time series from a structural causal process
    def lin_f(x): return x
    def nonlin_f(x): return (x + 5. * x**2 * np.exp(-x**2 / 20.))

    links = {0: [((0, -1), 0.9, lin_f)],
             1: [((1, -1), 0.8, lin_f), ((0, -1), 0.3, nonlin_f)],
             2: [((2, -1), 0.7, lin_f), ((1, 0), -0.2, lin_f)],
             }

    random_state_1 = np.random.default_rng(seed=1)
    random_state_2 = np.random.default_rng(seed=2)
    random_state_3 = np.random.default_rng(seed=3)

    noises = [random_state_1.standard_normal, random_state_2.standard_normal, random_state_3.standard_normal]

    ens = 3
    data_ens = {}
    for i in range(ens):
        data, nonstat = structural_causal_process(links,
                T=100, noises=noises)
        data[10, 1] == 999.
        data_ens[i] = data
    # print(data.shape)

    frame = DataFrame(data_ens, missing_flag=999.,
        analysis_mode = 'multiple')

    print(frame.T)

    # X=[(0, 0)]
    # Y=[(0, 0)]
    # Z=[(0, -3)]
    # tau_max=5
    # frame.construct_array(X, Y, Z, tau_max,
    #                     extraZ=None,
    #                     mask=None,
    #                     mask_type=None,
    #                     return_cleaned_xyz=False,
    #                     do_checks=True,
    #                     cut_off='2xtau_max',
    #                     verbosity=4)

    # print(ordinal_patt_array(data, array_mask=None, dim=2, step=1,
    #                     weights=False, verbosity=0)[0])